from itertools import product

from .datasets import DataSubject, DataCombination
from .flowchart import Edge
from .interface import PipelineOptions


class Step:

    def __init__(self, data_subject: DataSubject):
        self.data_subject = data_subject
        self.all_subjects = [data_subject]
        self._subgraphs = None

    def subgraphs(self, options: [PipelineOptions, None] = None):
        if options is None:
            options = PipelineOptions() #initialize with defaults

        if self._subgraphs is None:
            self._subgraphs = [subject.to_subgraph(options) for subject in self.all_subjects]
        return self._subgraphs

    def __repr__(self):
        return f'<Step(data_subject={self.data_subject})>'


class MergeStep(Step):

    def __init__(self, data_subject: DataSubject, merge_into_data_subject: DataSubject):
        self.merge_into_subject = merge_into_data_subject
        super().__init__(data_subject)
        self.all_subjects = [data_subject, merge_into_data_subject]
        self._combined_subject = None

    def subgraphs(self, options: [PipelineOptions, None] = None):
        if options is None:
            options = PipelineOptions() #initialize with defaults

        if self._subgraphs is None:
            subgraphs = []
            subgraphs.append(self._merge_subgraph(options))
            subgraphs.extend(super().subgraphs(options))
            self._subgraphs = subgraphs
        return self._subgraphs

    def _merge_subgraph(self, options: PipelineOptions):
        data_sources = []
        for orig_source, merge_source in product(self.data_subject.sources, self.merge_into_subject.sources):
            name = orig_source.name + ' ' + self.data_subject.name + '/' + merge_source.name + ' ' + self.merge_into_subject.name
            data_sources.append(DataCombination(orig_source, merge_source, name=name))

        subject_name = self.data_subject.name + '/' + self.merge_into_subject.name
        self._combined_subject = DataSubject(*data_sources, name=subject_name)
        return self._combined_subject.to_subgraph(options)

    @property
    def combined_subject(self):
        if self._combined_subject is None:
            self._merge_subgraph(PipelineOptions())
        return self._combined_subject

    def __repr__(self):
        return f'<MergeStep(data_subject={self.data_subject}, merge_subject={self.merge_into_subject})>'


class Process:

    def __init__(self, *steps: [Step]):
        self.steps = steps

    def __getitem__(self, item):
        return self.steps[item]

    def to_subgraphs(self, options: [PipelineOptions, None] = None):
        if options is None:
            options = PipelineOptions() #initialize with defaults

        subgraphs = []
        for step in self.steps:
            if isinstance(step, MergeStep):
                subgraphs.extend(step.subgraphs(options)) #contains last subgraph, combined subgraph, and to merge subgraph
            elif isinstance(step, Step):
                continue #step subgraph will be handled within merge step subgraph
            else:
                raise ValueError(f'must pass Step or MergeStep, got type {type(step)}')

        return subgraphs


    def to_edges(self, options: [PipelineOptions, None] = None, **edge_kwargs):
        if options is None:
            options = PipelineOptions() #initialize with defaults

        edges = []
        for i, step in enumerate(self.steps):
            if i == 0:
                continue #don't need to create edges with beginning step
            edges.extend(self._create_edges_for_step(step, options, **edge_kwargs))

        return edges

    def _create_edges_for_step(self, step: MergeStep, options: PipelineOptions, **edge_kwargs):
        # For current, need to split combined nodes and merge subgraphs
        combined_subgraph, last_subgraph, merge_subgraph = step.subgraphs(options)
        edges = []
        last_source_name, merge_source_name = _combined_source_name_to_individual_source_names(combined_subgraph.name)
        for combined_node in combined_subgraph.nodes:
            if options.data_sources: #extract node names by parsing combined name
                last_node_name, merge_node_name = _combined_name_to_input_and_merge_name(combined_node.name, last_source_name, merge_source_name)
            else: #now the source names are representative of the node names
                last_node_name = last_source_name
                merge_node_name = merge_source_name
            last_node = last_subgraph.nodes[last_node_name]
            merge_node = merge_subgraph.nodes[merge_node_name]

            if options.data_sources:
                edges.append(Edge(last_node, combined_node, **edge_kwargs))
                edges.append(Edge(merge_node, combined_node, **edge_kwargs))
            else:
                edges.append(Edge(last_node, combined_node, for_subgraphs=(last_subgraph, combined_subgraph)))
                edges.append(Edge(merge_node, combined_node, for_subgraphs=(merge_subgraph, combined_subgraph)))

        return edges

    def __repr__(self):
        return f'<Process(steps={self.steps})>'


def _to_combined_name(input_name, merge_name):
    return f'{input_name}/{merge_name}'

def _combined_name_to_input_and_merge_name(combined_name, last_source_name, merge_source_name):
    combined_name = combined_name.strip('DUMMY').replace(last_source_name, '').replace(merge_source_name, '')
    return _combined_source_name_to_individual_source_names(combined_name)


def _combined_source_name_to_individual_source_names(combined_name):
    if '/' not in combined_name:
        raise ValueError(f'passed regular name instead of combined name. combined name must have /. got {combined_name}')
    parts = combined_name.split('/')
    return '/'.join(parts[:-1]).strip(), parts[-1].strip()


from typing import Sequence

from datacode.models.transform.transform import Transform


def register_transforms(transforms: Sequence[Transform]):
    """
    Convenience function to add passed transforms into DEFAULT_TRANSFORMS which are automatically
    provided to all variables

    :param transforms:
    :return:
    """
    from datacode.models.transform.specific import DEFAULT_TRANSFORMS

    for transform in transforms:
        existing_keys = [transform.key for transform in DEFAULT_TRANSFORMS]
        if transform.key not in existing_keys:
            DEFAULT_TRANSFORMS.append(transform)



from typing import Sequence

import pandas as pd
from datacode.models.variables import Variable
from semopy import Optimizer

from datacode.sem.parse import lambda_df_from_opt


def get_latents(opt: Optimizer, mod_df: pd.DataFrame, latent_vars: Sequence[Variable]) -> pd.DataFrame:
    latent_df = _get_latents(opt, mod_df)
    rename_dict = {var.unique_key: var.name for var in latent_vars}
    return latent_df.rename(columns=rename_dict)


def _get_latents(opt: Optimizer, mod_df: pd.DataFrame) -> pd.DataFrame:
    model = opt.model
    lambda_df = lambda_df_from_opt(opt)
    for_mult_df = mod_df[lambda_df.index]
    factor_df = pd.DataFrame(
        for_mult_df.values @ lambda_df[model.vars['Latents']].values,
        columns=model.vars['Latents']
    )
    return factor_df



from typing import Sequence

from graphviz import Digraph

from datacode.graph.base import GraphObject


class Subgraph(GraphObject):

    def __init__(self, elems: Sequence[GraphObject], name=None, digraph=True, **graph_kwargs):
        self.elems = elems
        self.name = name
        self.digraph = digraph
        self.graph_kwargs = graph_kwargs

    def add_to_graph(self, graph: Digraph):
        full_name = 'cluster' + (self.name if self.name else '')
        subgraph = Digraph(full_name, **self.graph_kwargs)
        subgraph.attr('graph', label=self.name)
        [elem.add_to_graph(subgraph) for elem in self.elems]
        graph.subgraph(subgraph)

    def __repr__(self):
        return f'<Subgraph(name={self.name}, elems={self.elems})>'


from typing import Sequence, Tuple, Dict

import more_itertools
import pandas as pd
import matplotlib.pyplot as plt


def variables_drop_count_df_and_plt_figs(
    df: pd.DataFrame, var_names: Sequence[str], id_var: str
) -> Tuple[pd.DataFrame, Dict[str, plt.Figure]]:
    for_concat = []
    non_missing_counts = df[var_names].count()
    non_missing_counts.name = 'Count'
    for_concat.append(non_missing_counts)
    if id_var is not None:
        by_id_counts = _by_id_counts(df, var_names, id_var)
        by_id_counts.name = id_var
        for_concat.append(by_id_counts)

    count_with_all = len(df.dropna(subset=var_names))

    detail_df = pd.DataFrame(index=var_names, columns=['Min', 'Average', 'Median', 'Max', 'Mode'])
    figs: Dict[str, plt.Figure] = {}
    for var_name in var_names:
        _, hist_fig, min_, avg, median, max_, mode = _variable_drop_count_details(df, var_name, var_names,
                                                                                  count_with_all)
        detail_df.loc[var_name, 'Min'] = min_
        detail_df.loc[var_name, 'Average'] = avg
        detail_df.loc[var_name, 'Median'] = median
        detail_df.loc[var_name, 'Max'] = max_
        detail_df.loc[var_name, 'Mode'] = mode
        figs[var_name] = hist_fig

    out_df = pd.concat([*for_concat, detail_df], axis=1)
    return out_df, figs


def _by_id_counts(df: pd.DataFrame, var_names: Sequence[str], id_var: str) -> pd.Series:
    id_counts = df.groupby(id_var)[var_names].count()
    return (id_counts > 0).sum()


def _variable_drop_count_details(df: pd.DataFrame, var_name: str, var_names: Sequence[str],
                                 count_with_all: int
                                 ) -> Tuple[Dict[Sequence[str], int], plt.Figure, float, float, float, float, float]:
    other_var_names = [name for name in var_names if name != var_name]
    count_with_all_but_var = len(df.dropna(subset=other_var_names))
    gain_from_drop = count_with_all_but_var - count_with_all

    gains = _gain_from_drop_combos(df, var_name, var_names)

    gains_df = pd.DataFrame()
    gains_df['Counts'] = gains.values()
    hist_fig = _gains_histogram(gains_df)

    min_ = gains_df['Counts'].min()
    avg = gains_df['Counts'].mean()
    median = gains_df['Counts'].median()
    max_ = gains_df['Counts'].max()
    mode = gains_df['Counts'].mode().values[0]

    return gains, hist_fig, min_, avg, median, max_, mode


def _gain_from_drop_combos(df: pd.DataFrame, var_name: str, var_names: Sequence[str]):
    other_var_names = [name for name in var_names if name != var_name]
    gains = {}
    for other_var_combo in more_itertools.powerset(other_var_names):
        if not other_var_combo:
            continue
        full_combo = other_var_combo + (var_name,)
        count = len(df.dropna(subset=full_combo))
        count_without = len(df.dropna(subset=other_var_combo))
        gain = count_without - count
        gains[other_var_combo] = gain
    return gains


def _gains_histogram(gains_df: pd.DataFrame) -> plt.Figure:
    ax = gains_df['Counts'].hist()
    fig = ax.get_figure()
    plt.close()

    return fig


from pandas.testing import assert_frame_equal

from datacode import DataSource
from datacode.models.pipeline.operations.combine import CombineOptions
from datacode.models.pipeline.operations.merge import MergeOptions
from tests.pipeline.base import PipelineTest


class TestDataCombinationPipeline(PipelineTest):

    def test_create_and_run_combine_rows_pipeline_from_sources(self):
        dp = self.create_combine_pipeline()
        dp.execute()

        assert_frame_equal(dp.df, self.expect_combined_rows_1_2)

    def test_create_and_run_combine_rows_drop_rows_pipeline_from_sources(self):
        a, b, c = self.create_variables()
        co = CombineOptions(row_duplicate_vars=[c])
        dp = self.create_combine_pipeline(combine_options_list=[co])
        dp.execute()

        assert_frame_equal(dp.df, self.expect_combined_rows_1_2_row_drop_c)

    def test_create_and_run_combine_rows_drop_entities_pipeline_from_sources(self):
        a, b, c = self.create_variables()
        co = CombineOptions(entity_duplicate_vars=[c])
        dp = self.create_combine_pipeline(combine_options_list=[co])
        dp.execute()

        assert_frame_equal(dp.df, self.expect_combined_rows_1_2_entity_drop_c)

    def test_create_and_run_combine_cols_pipeline_from_sources(self):
        co = CombineOptions(rows=False)
        dp = self.create_combine_pipeline(combine_options_list=[co])

        with self.assertRaises(ValueError) as cm:
            dp.execute()
            exc = cm.exception
            assert 'exists in multiple data sources' in str(exc)

    def test_create_and_run_combine_rows_pipeline_with_indices(self):
        dp = self.create_combine_pipeline(indexed=True)
        dp.execute()

        assert_frame_equal(dp.df, self.expect_combined_rows_1_2_c_index)

    def test_create_and_run_combine_cols_pipeline_with_indices(self):
        co = CombineOptions(rows=False)
        dp = self.create_combine_pipeline(indexed=True, combine_options_list=[co])
        dp.execute()

        assert_frame_equal(dp.df, self.expect_merged_1_2_c_index)

    def test_auto_run_pipeline_by_load_source_with_no_location(self):
        dp = self.create_combine_pipeline()

        ds = DataSource(pipeline=dp, location=self.csv_path_output)
        df = ds.df
        assert_frame_equal(df, self.expect_combined_rows_1_2)

    def test_create_and_run_combine_pipeline_three_sources(self):
        dp = self.create_combine_pipeline(include_indices=(0, 1, 2))
        dp.execute()

        assert_frame_equal(dp.df, self.expect_combined_rows_1_2_3)

    def test_raises_error_for_mismatching_data_sources_merge_options(self):
        co = CombineOptions()
        dp = self.create_combine_pipeline(include_indices=(0, 1, 2), combine_options_list=[co])

        with self.assertRaises(ValueError) as cm:
            dp.execute()
            exc = cm.exception
            assert 'must have one fewer combine options than data sources' in str(exc)

    def test_create_nested_pipeline(self):
        dp1 = self.create_combine_pipeline(include_indices=(0, 1))

        self.create_csv_for_3()
        ds3_cols = self.create_columns_for_3()
        ds3 = self.create_source(df=None, location=self.csv_path3, columns=ds3_cols, name='three')

        dp2 = self.create_combine_pipeline(data_sources=[dp1, ds3])
        dp2.execute()

        assert_frame_equal(dp2.df, self.expect_combined_rows_1_2_3)




import os
from typing import Optional, Tuple, Sequence, List

import pandas as pd
from numpy import nan

from datacode import AnalysisOptions, DataSource, DataAnalysisPipeline, DataGeneratorPipeline, GenerationOptions, \
    Variable, Column, MergeOptions, DataMergePipeline, SourceTransform, DataTransformationPipeline, TransformOptions, \
    ColumnIndex
from datacode.models.pipeline.combine import DataCombinationPipeline
from datacode.models.pipeline.operations.combine import CombineOptions
from datacode.models.types import DataSourceOrPipeline
from tests.test_source import SourceTest
from tests.utils import GENERATED_PATH


def analysis_from_source(ds: DataSource) -> float:
    running_sum = 0
    for var in ds.load_variables:
        if var.dtype.is_numeric:
            running_sum += ds.df[var.name].sum()
    return running_sum


EXPECT_GENERATED_DF = df = pd.DataFrame(
    [
        (1, 2, 'd'),
        (3, 4, 'e')
    ],
    columns=['a', 'b', 'C']
).convert_dtypes()


def ds_generator_func(columns: Sequence[Column]) -> DataSource:
    ds = DataSource(df=EXPECT_GENERATED_DF, columns=columns)
    return ds


def source_transform_func(ds: DataSource) -> DataSource:
    for variable in ds.load_variables:
        if variable.dtype.is_numeric:
            ds.df[variable.name] += 1
    return ds


class PipelineTest(SourceTest):
    merge_var = Variable('c', 'C', dtype='str')
    source_transform = SourceTransform('st', name_func=SourceTest.transform_name_func, data_func=source_transform_func)
    ds_one_analysis_result = 21
    ds_one_and_two_analysis_result = 191
    ds_one_transformed_analysis_result = 27
    ds_one_generated_analysis_result = 10
    csv_path2 = os.path.join(GENERATED_PATH, 'data2.csv')
    csv_path3 = os.path.join(GENERATED_PATH, 'data3.csv')
    csv_path_output = os.path.join(GENERATED_PATH, 'output.csv')

    test_df2 = pd.DataFrame(
        [
            (10, 20, 'd'),
            (50, 60, 'e'),
        ],
        columns=['e', 'f', 'c']
    )
    test_df3 = pd.DataFrame(
        [
            (100, 200, 'd'),
            (500, 600, 'e'),
        ],
        columns=['g', 'h', 'c']
    )
    expect_merged_1_2 = pd.DataFrame(
        [
            (1, 2, 'd', 10, 20),
            (3, 4, 'd', 10, 20),
            (5, 6, 'e', 50, 60),
        ],
        columns=['A', 'B', 'C', 'E', 'F']
    ).convert_dtypes()
    expect_merged_1_2_c_index = expect_merged_1_2.set_index('C')
    expect_merged_1_transformed_2 = pd.DataFrame(
        [
            (2, 3, 'd', 10, 20),
            (4, 5, 'd', 10, 20),
            (6, 7, 'e', 50, 60),
        ],
        columns=['A', 'B', 'C', 'E', 'F']
    ).convert_dtypes()
    expect_merged_1_2_both_transformed = pd.DataFrame(
        [
            (2, 3, 'd', 11, 21),
            (4, 5, 'd', 11, 21),
            (6, 7, 'e', 51, 61),
        ],
        columns=['A', 'B', 'C', 'E', 'F']
    ).convert_dtypes()
    expect_merged_1_generated_2 = pd.DataFrame(
        [
            (1, 2, 'd', 10, 20),
            (3, 4, 'e', 50, 60),
        ],
        columns=['a', 'b', 'C', 'E', 'F']
    ).convert_dtypes()
    expect_merged_1_2_3 = pd.DataFrame(
        [
            (1, 2, 'd', 10, 20, 100, 200),
            (3, 4, 'd', 10, 20, 100, 200),
            (5, 6, 'e', 50, 60, 500, 600),
        ],
        columns=['A', 'B', 'C', 'E', 'F', 'G', 'H']
    ).convert_dtypes()
    expect_func_df = df = pd.DataFrame(
        [
            (2, 3, 'd'),
            (4, 5, 'd'),
            (6, 7, 'e')
        ],
        columns=['A', 'B', 'C']
    ).convert_dtypes()
    expect_func_df_with_a_and_a_transformed = pd.DataFrame(
        [
            (2, 3, 3, 'd'),
            (4, 5, 5, 'd'),
            (6, 7, 7, 'e')
        ],
        columns=['A', 'A_1', 'B', 'C']
    ).convert_dtypes()
    expect_loaded_df_with_transform = pd.DataFrame(
        [
            (2, 3, 'd'),
            (4, 5, 'd'),
            (6, 7, 'e')
        ],
        columns=['A_1', 'B_1', 'C_1']
    ).convert_dtypes()
    expect_df_double_source_transform = pd.DataFrame(
        [
            (3, 4, 'd'),
            (5, 6, 'd'),
            (7, 8, 'e')
        ],
        columns=['A', 'B', 'C']
    ).convert_dtypes()
    expect_generated_transformed = pd.DataFrame(
        [
            (2, 3, 'd'),
            (4, 5, 'e')
        ],
        columns=['a', 'b', 'C']
    ).convert_dtypes()
    expect_combined_rows_1_2 = pd.DataFrame(
        [
            (1, 2, 'd', nan, nan),
            (3, 4, 'd', nan, nan),
            (5, 6, 'e', nan, nan),
            (nan, nan, 'd', 10, 20),
            (nan, nan, 'e', 50, 60),
        ],
        columns=['A', 'B', 'C', 'E', 'F'],
    ).convert_dtypes()
    expect_combined_rows_1_2_c_index = expect_combined_rows_1_2.set_index('C')
    expect_combined_rows_1_2_3 = pd.DataFrame(
        [
            (1, 2, 'd', nan, nan, nan, nan),
            (3, 4, 'd', nan, nan, nan, nan),
            (5, 6, 'e', nan, nan, nan, nan),
            (nan, nan, 'd', 10, 20, nan, nan),
            (nan, nan, 'e', 50, 60, nan, nan),
            (nan, nan, 'd', nan, nan, 100, 200),
            (nan, nan, 'e', nan, nan, 500, 600),
        ],
        columns=['A', 'B', 'C', 'E', 'F', 'G', 'H']
    ).convert_dtypes()
    expect_combined_rows_1_2_entity_drop_c = pd.DataFrame(
        [
            (1, 2, 'd', nan, nan),
            (3, 4, 'd', nan, nan),
            (5, 6, 'e', nan, nan),
        ],
        columns=['A', 'B', 'C', 'E', 'F']
    ).convert_dtypes()
    expect_combined_rows_1_2_row_drop_c = pd.DataFrame(
        [
            (1, 2, 'd', nan, nan),
            (5, 6, 'e', nan, nan),
        ],
        columns=['A', 'B', 'C', 'E', 'F']
    ).convert_dtypes()





    def create_csv_for_2(self, df: Optional[pd.DataFrame] = None, **to_csv_kwargs):
        if df is None:
            df = self.test_df2
        df.to_csv(self.csv_path2, index=False, **to_csv_kwargs)

    def create_csv_for_3(self, df: Optional[pd.DataFrame] = None, **to_csv_kwargs):
        if df is None:
            df = self.test_df3
        df.to_csv(self.csv_path3, index=False, **to_csv_kwargs)

    def create_variables_for_2(self, transform_data: str = '', apply_transforms: bool = True
                               ) -> Tuple[Variable, Variable, Variable]:
        transform_dict = self.get_transform_dict(transform_data=transform_data, apply_transforms=apply_transforms)

        e = Variable('e', 'E', dtype='int', **transform_dict)
        f = Variable('f', 'F', dtype='int', **transform_dict)
        c = Variable('c', 'C', dtype='str')
        return e, f, c

    def create_variables_for_3(self, transform_data: str = '', apply_transforms: bool = True
                               ) -> Tuple[Variable, Variable, Variable]:
        transform_dict = self.get_transform_dict(transform_data=transform_data, apply_transforms=apply_transforms)

        g = Variable('g', 'G', dtype='int', **transform_dict)
        h = Variable('h', 'H', dtype='int', **transform_dict)
        c = Variable('c', 'C', dtype='str')
        return g, h, c

    def create_variables_for_generated(self, transform_data: str = '', apply_transforms: bool = True
                               ) -> Tuple[Variable, Variable, Variable]:
        transform_dict = self.get_transform_dict(transform_data=transform_data, apply_transforms=apply_transforms)

        a = Variable('a', 'a', dtype='int', **transform_dict)
        b = Variable('b', 'b', dtype='int', **transform_dict)
        c = Variable('c', 'C', dtype='str')
        return a, b, c

    def create_columns_for_2(self, transform_data: str = '', apply_transforms: bool = True):
        e, f, c = self.create_variables_for_2(transform_data=transform_data, apply_transforms=apply_transforms)
        ec = Column(e, 'e')
        fc = Column(f, 'f')
        cc = Column(c, 'c')
        return [
            ec,
            fc,
            cc
        ]

    def create_columns_for_3(self, transform_data: str = '', apply_transforms: bool = True):
        g, h, c = self.create_variables_for_3(transform_data=transform_data, apply_transforms=apply_transforms)
        gc = Column(g, 'g')
        hc = Column(h, 'h')
        cc = Column(c, 'c')
        return [
            gc,
            hc,
            cc
        ]

    def create_columns_for_generated(self, transform_data: str = '', apply_transforms: bool = True):
        a, b, c = self.create_variables_for_generated(transform_data=transform_data, apply_transforms=apply_transforms)
        ac = Column(a, 'a')
        bc = Column(b, 'b')
        cc = Column(c, 'C')
        return [
            ac,
            bc,
            cc
        ]

    def create_variables_and_c_colindex_for_2(self, transform_data: str = '', apply_transforms: bool = True
                                        ) -> Tuple[List[Variable], ColumnIndex]:
        e, f, c = self.create_variables_for_2(transform_data=transform_data, apply_transforms=apply_transforms)
        c_index = self.create_c_index()

        c_col_index = ColumnIndex(c_index, [c])

        return [e, f, c], c_col_index

    def create_indexed_columns_for_2(self, transform_data: str = '', apply_transforms: bool = True) -> List[Column]:
        (e, f, c), c_col_index = self.create_variables_and_c_colindex_for_2(
            transform_data=transform_data, apply_transforms=apply_transforms
        )
        ec = Column(e, 'e', indices=[c_col_index])
        fc = Column(f, 'f', indices=[c_col_index])
        cc = Column(c, 'c')
        return [
            ec,
            fc,
            cc
        ]

    def create_variables_and_c_colindex_for_3(self, transform_data: str = '', apply_transforms: bool = True
                                        ) -> Tuple[List[Variable], ColumnIndex]:
        g, h, c = self.create_variables_for_3(transform_data=transform_data, apply_transforms=apply_transforms)
        c_index = self.create_c_index()

        c_col_index = ColumnIndex(c_index, [c])

        return [g, h, c], c_col_index

    def create_indexed_columns_for_3(self, transform_data: str = '', apply_transforms: bool = True) -> List[Column]:
        (g, h, c), c_col_index = self.create_variables_and_c_colindex_for_3(
            transform_data=transform_data, apply_transforms=apply_transforms
        )
        gc = Column(g, 'g', indices=[c_col_index])
        hc = Column(h, 'h', indices=[c_col_index])
        cc = Column(c, 'c')
        return [
            gc,
            hc,
            cc
        ]

    def create_merge_pipeline(self, include_indices: Sequence[int] = (0, 1),
                              data_sources: Optional[Sequence[DataSource]] = None,
                              merge_options_list: Optional[Sequence[MergeOptions]] = None,
                              indexed: bool = False) -> DataMergePipeline:
        if indexed:
            col_func_1 = self.create_indexed_columns
            col_func_2 = self.create_indexed_columns_for_2
            col_func_3 = self.create_indexed_columns_for_3
        else:
            col_func_1 = self.create_columns
            col_func_2 = self.create_columns_for_2
            col_func_3 = self.create_columns_for_3


        if data_sources is None:
            self.create_csv()
            ds1_cols = col_func_1()
            ds1 = self.create_source(df=None, columns=ds1_cols, name='one')
            self.create_csv_for_2()
            ds2_cols = col_func_2()
            ds2 = self.create_source(df=None, location=self.csv_path2, columns=ds2_cols, name='two')
            self.create_csv_for_3()
            ds3_cols = col_func_3()
            ds3 = self.create_source(df=None, location=self.csv_path3, columns=ds3_cols, name='three')


            data_sources = [ds1, ds2, ds3]
            selected_data_sources = []
            for i, ds in enumerate(data_sources):
                if i in include_indices:
                    selected_data_sources.append(ds)
        else:
            selected_data_sources = data_sources

        if merge_options_list is None:
            mo = MergeOptions([self.merge_var.name])
            merge_options_list = [mo for _ in range(len(selected_data_sources) - 1)]

        merge_options_list[-1].out_path = self.csv_path_output

        dp = DataMergePipeline(selected_data_sources, merge_options_list)
        return dp


    def create_analysis_pipeline(self, source: Optional[DataSourceOrPipeline] = None,
                                 options: Optional[AnalysisOptions] = None):
        if source is None:
            self.create_csv()
            ds1_cols = self.create_columns()
            source = self.create_source(df=None, columns=ds1_cols, name='one')

        if options is None:
            options = AnalysisOptions(analysis_from_source)

        dap = DataAnalysisPipeline(source, options)
        return dap

    def create_generator_pipeline(self, **kwargs) -> DataGeneratorPipeline:
        gen_cols = self.create_columns_for_generated()
        config_dict = dict(
            out_path=self.csv_path_output, columns=gen_cols
        )
        config_dict.update(**kwargs)
        go = GenerationOptions(ds_generator_func, **config_dict)
        dgp = DataGeneratorPipeline(go)
        return dgp

    def create_transformation_pipeline(self, source: Optional[DataSourceOrPipeline] = None,
                                       **pipeline_kwargs) -> DataTransformationPipeline:
        config_dict = dict(
            func=source_transform_func,
            out_path=self.csv_path_output
        )
        config_dict.update(pipeline_kwargs)
        if source is None:
            self.create_csv()
            all_cols = self.create_columns()
            source = self.create_source(df=None, columns=all_cols)

        to = TransformOptions(**config_dict)

        dtp = DataTransformationPipeline(source, to)
        return dtp

    def create_combine_pipeline(self, include_indices: Sequence[int] = (0, 1),
                                data_sources: Optional[Sequence[DataSource]] = None,
                                combine_options_list: Optional[Sequence[CombineOptions]] = None,
                                indexed: bool = False):
        if indexed:
            col_func_1 = self.create_indexed_columns
            col_func_2 = self.create_indexed_columns_for_2
            col_func_3 = self.create_indexed_columns_for_3
        else:
            col_func_1 = self.create_columns
            col_func_2 = self.create_columns_for_2
            col_func_3 = self.create_columns_for_3


        if data_sources is None:
            self.create_csv()
            ds1_cols = col_func_1()
            ds1 = self.create_source(df=None, columns=ds1_cols, name='one')
            self.create_csv_for_2()
            ds2_cols = col_func_2()
            ds2 = self.create_source(df=None, location=self.csv_path2, columns=ds2_cols, name='two')
            self.create_csv_for_3()
            ds3_cols = col_func_3()
            ds3 = self.create_source(df=None, location=self.csv_path3, columns=ds3_cols, name='three')


            data_sources = [ds1, ds2, ds3]
            selected_data_sources = []
            for i, ds in enumerate(data_sources):
                if i in include_indices:
                    selected_data_sources.append(ds)
        else:
            selected_data_sources = data_sources

        if combine_options_list is None:
            mo = CombineOptions()
            combine_options_list = [mo for _ in range(len(selected_data_sources) - 1)]

        combine_options_list[-1].out_path = self.csv_path_output

        dp = DataCombinationPipeline(selected_data_sources, combine_options_list)
        return dp


import pandas as pd

def _how_merge_df(df: pd.DataFrame, other_df: pd.DataFrame, ids, how='left'):
    return df.merge(other_df, on=ids, how=how)

def outer_merge_df(df: pd.DataFrame, other_df: pd.DataFrame, ids):
    return _how_merge_df(df, other_df, ids, how='outer')

def left_merge_df(df: pd.DataFrame, other_df: pd.DataFrame, ids):
    return _how_merge_df(df, other_df, ids, how='left')

def right_merge_df(df: pd.DataFrame, other_df: pd.DataFrame, ids):
    return _how_merge_df(df, other_df, ids, how='right')


def _get_obj_or_attr(obj, obj_or_attr):
    if obj_or_attr == 'obj':
        return obj
    else:
        return getattr(obj, obj_or_attr)


def _to_list_if_str(list_or_str):
    if isinstance(list_or_str, str):
        return [list_or_str]
    return list_or_str


"""
A framework to manage variables. Annotates data such that transformations of a variable are seen as that and not
as new variables. Manages naming the variable across transformations.
"""
from .collection import (
    VariableCollection,
    VariableDisplayNameCollection,
    VariableNameCollection,
    Variable,
    VariableChangeNameCollection,
    VariablePortfolioNameCollection,
    VariableChangePortfolioNameCollection
)


from typing import Optional, Union

from sympy import Symbol


def to_symbol_if_necessary(sym: Optional[Union[str, Symbol]]) -> Optional[Symbol]:
    if isinstance(sym, Symbol) or sym is None:
        return sym

    return str_to_symbol(sym)


def str_to_symbol(sym: str) -> Symbol:
    return Symbol(sym)


def var_key_to_symbol_str(var_key: str) -> str:
    parts = var_key.split('_')
    if len(parts) == 1:
        return rf'\text{{{parts[0].title()}}}'
    elif len(parts) > 1:
        return "".join([part[0].upper() for part in parts])
    else:
        raise ValueError(f'got invalid var_key {var_key}')



from datacode.ff.fftypes import (
    DictofStrsandStrLists,
    GroupvarNgroupsDict,
    TwoStrTupleList,
    StrBoolDict,
    Tuple
)
from datacode.ff.create.sort import _other_groupvar_portname

def _standardize_custom_args(custom_labels: DictofStrsandStrLists = None,
                             custom_groupvar_ngroups_dict: GroupvarNgroupsDict = None,
                             custom_pairings: TwoStrTupleList = None,
                             custom_low_minus_high_dict: StrBoolDict = None
                             ) -> Tuple[DictofStrsandStrLists, GroupvarNgroupsDict, TwoStrTupleList, StrBoolDict]:

    # Replace variables with portfolio variables
    if custom_labels is not None:
        custom_labels = {_other_groupvar_portname(key): value for key, value in custom_labels.items()}

    if custom_low_minus_high_dict is not None:
        custom_low_minus_high_dict = {_other_groupvar_portname(key): value for key, value in custom_low_minus_high_dict.items()}

    if custom_pairings is not None:
        new_custom_pairings = []
        for pairing in custom_pairings:
            new_custom_pairings.append(
                (
                    _other_groupvar_portname(pairing[0]),
                    _other_groupvar_portname(pairing[1])
                )
            )
    else:
        new_custom_pairings = None

    return custom_labels, custom_groupvar_ngroups_dict, new_custom_pairings, custom_low_minus_high_dict


import operator
from typing import TYPE_CHECKING, Sequence, Callable, Optional

from mixins import ReprMixin
from sympy import Expr, lambdify, latex
import pandas as pd

from datacode.models.variables.compare import functions_are_equal

if TYPE_CHECKING:
    from datacode.models.variables.variable import Variable
    from datacode.models.column.column import Column


class Expression(ReprMixin):
    repr_cols = ['explanation', 'variables']

    def __init__(self, variables: Sequence['Variable'], func: Callable[[Sequence['Column']], pd.Series],
                 expr: Optional[Expr] = None, summary: Optional[str] = None):
        self.variables = variables
        self.func = func
        self.expr = expr
        self.summary = summary

    @classmethod
    def from_sympy_expr(cls, variables: Sequence['Variable'], expr: Expr, **kwargs):
        def eval_sympy_eq(cols: Sequence['Column']) -> pd.Series:
            symbols = []
            data = []
            for col in cols:
                if col.series is None:
                    raise ValueError(f'Need column {col} to have data for calculation, but col.series is None')
                symbols.append(col.variable.symbol)
                data.append(col.series)

            func = lambdify(symbols, expr)
            series = func(*data)
            return series

        return cls(
            variables,
            eval_sympy_eq,
            expr=expr
        )

    @property
    def explanation(self) -> Optional[str]:
        if self.summary is not None:
            return self.summary

        if self.expr is not None:
            return latex(self.expr)

    def __eq__(self, other):
        same = True
        try:
            same = same and self.variables == other.variables
            same = same and functions_are_equal(self.func, other.func)
            same = same and self.expr == other.expr
            same = same and self.summary == other.summary
            return same
        except AttributeError:
            return False

    def __add__(self, other):
        return self._create_expression_from_other_and_operator(other, operator.add, 'add')

    def __sub__(self, other):
        return self._create_expression_from_other_and_operator(other, operator.sub, 'subtract', preposition='from')

    def __mul__(self, other):
        return self._create_expression_from_other_and_operator(other, operator.mul, 'multiply', preposition='by')

    def __truediv__(self, other):
        return self._create_expression_from_other_and_operator(other, operator.truediv, 'divide', preposition='by')

    def __pow__(self, other):
        return self._create_expression_from_other_and_operator(other, operator.pow, 'exponentiate', preposition='by')

    def _create_expression_from_other_and_operator(self, other, op_func: Callable, operator_name: str,
                                                   preposition: str = 'to') -> 'Expression':
        from datacode.models.variables.variable import Variable

        if self.expr is None:
            raise ValueError(f'cannot {operator_name} expression which does not have .expr, got {self}')

        if isinstance(other, Variable):
            sympy_expr = op_func(self.expr, other.symbol)
            expr = Expression.from_sympy_expr([*self.variables, other], sympy_expr)
            return expr

        if isinstance(other, Expression):
            if other.expr is None:
                raise ValueError(f'cannot {operator_name} expression which does not have .expr, got {other}')
            sympy_expr = op_func(self.expr, other.expr)
            expr = Expression.from_sympy_expr([*self.variables, *other.variables], sympy_expr)
            return expr

        raise ValueError(f'Cannot {operator_name} {other} of type {type(other)} {preposition} {self}, '
                         f'must be Variable or Expression')



import pandas as pd
import statsmodels.api as sm
from regtools.iter import reg_for_each_xvar_set_and_produce_summary
from datacode.psm.typing import StrList, StrListOrNone, RegResultSummaryTuple
from datacode.psm.names import prob_treated_varname, predict_treated_varname, accurate_prediction_name


def explain_probability_of_treatment(df: pd.DataFrame, treated_var: str, xvars: StrList, fe: StrListOrNone = None
                                     ) -> RegResultSummaryTuple:
    reg_list, summ = reg_for_each_xvar_set_and_produce_summary(
        df,
        treated_var,
        [xvars],
        stderr=True,
        fe=fe,
        model_type='logit'
    )

    return reg_list[0], summ

def predict_probability_of_treatment(df: pd.DataFrame, treated_var: str, xvars: StrList,
                                     entity_var: str, time_var: str,
                                     fe: StrListOrNone = None
                                     ) -> pd.DataFrame:

    result, summ = explain_probability_of_treatment(
        df,
        treated_var,
        xvars,
        fe=fe
    )

    X = sm.add_constant(df[xvars])

    prob_treated_var = prob_treated_varname(treated_var)
    predict_treated_var = predict_treated_varname(treated_var)

    predict_df = df[[entity_var, time_var, treated_var]]
    predict_df[prob_treated_var] = result.predict(X)
    predict_df[predict_treated_var] = predict_df[prob_treated_var].apply(lambda x: 1 if x >= 0.5 else 0)
    predict_df[accurate_prediction_name] = predict_df[predict_treated_var] == predict_df[treated_var]

    return predict_df



import pandas as pd

from datacode.summarize.subset.outliers.summary.overview import bad_df_summary_df
from datacode.summarize.subset.outliers.summary.stats import full_df_summary_stats_df
from datacode.summarize.subset.outliers.select import drop_outliers_by_cutoffs
from datacode.summarize.subset.outliers.summary.totex import df_dict_to_table
from datacode.summarize.subset.outliers.typing import (
    MinMaxDict,
    FloatSequence
)
from datacode.typing import DocumentOrTables


def outlier_overview_summary_page_table(df: pd.DataFrame, bad_df: pd.DataFrame,
                                        min_max_dict: MinMaxDict,
                                        bad_column_name: str = '_bad_column',
                                        firm_id_col: str = 'TICKER',
                                        outdir: str = None, as_document=True, author: str = None,
                                        caption: str='Outlier Overview and Summary Statistics',
                                        quants: FloatSequence=(0.01, 0.1, 0.5, 0.9, 0.99)
                                        ) -> DocumentOrTables:

    outlier_cols = list(min_max_dict.keys())
    valid_df = drop_outliers_by_cutoffs(df, min_max_dict)

    overview_df = bad_df_summary_df(
        bad_df,
        min_max_dict,
        firm_id_col=firm_id_col,
        bad_column_name=bad_column_name
    )

    outlier_stats, full_stats, no_outlier_stats = [full_df_summary_stats_df(
        current_df,
        outlier_cols,
        quants=quants
    ) for current_df in (bad_df, df, valid_df)]

    panel_name_df_dict = {
        'Outlier Overview by Variable': overview_df,
        'Summary Statistics - Outlier Observations': outlier_stats,
        'Summary Statistics - Full Sample': full_stats,
        'Summary Statistics - Trimmed Sample': no_outlier_stats
    }

    document_or_table = df_dict_to_table(
        panel_name_df_dict,
        as_document=as_document,
        author=author,
        caption=caption,
    )

    if outdir:
        document_or_table.to_pdf_and_move(
            outdir,
            outname=caption
        )

    return document_or_table






import io
from contextlib import redirect_stdout

import pandas as pd
import numpy as np
from semopy import Model, Optimizer

from datacode.sem.exc import SampleCovMatrixNotPositiveDefiniteException, ModelCovMatrixNotPositiveDefiniteException


def run_model(model_def: str, df: pd.DataFrame, pos_def_warnings: str = 'ignore',
              **opt_kwargs) -> Optimizer:
    pos_def_warnings = _validate_pos_def(pos_def_warnings)
    model = Model(model_def)
    model.load_dataset(df)

    if pos_def_warnings == 'error':
        try:
            np.linalg.cholesky(model.mx_cov)
        except np.linalg.LinAlgError:
            raise SampleCovMatrixNotPositiveDefiniteException

    with io.StringIO() as buf, redirect_stdout(buf):
        opt = Optimizer(model)
        opt.optimize(**opt_kwargs)
        stdout_val = buf.getvalue()

    if pos_def_warnings == 'error':
        sigma = opt.get_sigma()[0]
        try:
            np.linalg.cholesky(sigma)
        except np.linalg.LinAlgError:
            raise ModelCovMatrixNotPositiveDefiniteException

    if pos_def_warnings == 'print':
        print(stdout_val)

    opt._sample_cov_pos_df = True
    opt._model_cov_pos_df = True
    if 'a sample covariance matrix is not positive-definite' in stdout_val:
        opt._sample_cov_pos_df = False
    if 'resulting model-implied covariance matrix is not postive-definite' in stdout_val:
        opt._model_cov_pos_df = False

    return opt


def _validate_pos_def(pos_def_str: str) -> str:
    pos_def_str = pos_def_str.casefold().strip()
    if pos_def_str not in ('print', 'ignore', 'error'):
        raise ValueError('invalid positive definite string, pass one of print, ignore, or error')
    return pos_def_str



from typing import Sequence

import pandas as pd
from IPython.core.display import display, HTML

from datacode.typing import DfDict

def display_df_dict(df_dict: DfDict):
    # df_dict may either be one level deep or two levels deep.
    # if two levels deep, then first level organizes which table it is,
    # and the second level is panels
    for df_name in df_dict:
        df_or_dict_or_dfs = df_dict[df_name]
        # Handle the case where it is two levels deep
        if isinstance(df_or_dict_or_dfs, dict):
            display(_html_from_str(df_name))
            display_df_dict(df_or_dict_or_dfs)
        elif isinstance(df_or_dict_or_dfs, (list, tuple)):
            display_dfs(df_or_dict_or_dfs, df_name)
        else:
            _display_df(df_or_dict_or_dfs, df_name)


def display_dfs(df_list: Sequence[pd.DataFrame], title: str=None) -> None:

    if title is not None:
        display(_html_from_str(title, tag_type='h3'))

    display(*df_list)


def _display_df(df: pd.DataFrame, df_name: str):
    display(_html_from_str(df_name, tag_type='h3'))
    display(df)


def _html_from_str(str_: str, tag_type='h2') -> HTML:
    return HTML(f'<{tag_type}>{str_}</{tag_type}>')


from copy import deepcopy
from typing import Type

import numpy as np

from datacode.models.dtypes.base import DataType
from datacode.models.dtypes.bit_size import get_bit_from_dtype


class FloatType(DataType):
    names = ('float', 'decimal', 'float16', 'float32', 'float64', 'float128', 'floating')

    def __init__(self, categorical: bool = False, ordered: bool = False, bit_size: int = 64):
        self.bit_size = bit_size
        super().__init__(
            float,
            pd_class=self._get_pd_class(),
            categorical=categorical,
            ordered=ordered,
            is_numeric=True
        )
        self.equal_attrs = deepcopy(self.equal_attrs)
        self.equal_attrs.append('bit_size')
        self.repr_cols = deepcopy(self.repr_cols)
        self.repr_cols.append('bit_size')

    def _get_pd_class(self) -> Type:
        if self.bit_size == 16:
            return np.float16
        elif self.bit_size == 32:
            return np.float32
        elif self.bit_size == 64:
            return np.float64
        elif self.bit_size == 128:
            return np.float128
        else:
            raise ValueError(f'must pass bit_size of 16, 32, 64 or 128. Got {self.bit_size}')

    @classmethod
    def from_str(cls, dtype: str, categorical: bool = False, ordered: bool = False):
        dtype = dtype.lower()
        if dtype not in cls.names:
            raise ValueError(f'Dtype {dtype} does not match valid names for {cls.__name__}: {cls.names}')
        bit_or_none = get_bit_from_dtype(dtype)
        if bit_or_none is None:
            bit_or_none = 64
        return cls(
            categorical=categorical,
            ordered=ordered,
            bit_size=bit_or_none
        )



import pandas as pd
import os

from ..tools import _ids_str_or_tuple_to_id_name
from .flowchart import Node, Subgraph
from .interface import PipelineOptions, IDSet, IDCollection
from ..backend import compare_id_collections
from .bars import MatchComparisonBarGraph


class DataItem:

    def __init__(self, name):
        self.name = name


class DataSource(DataItem):

    def __init__(self, name=None, id_cols=None, data_source=None, data_loader=None, **data_loader_kwargs):
        """

        :param name:
        :type name:
        :param id_cols: list of tuples or strs of id sets to match on, e.g. ['CUSIP6', ('CUSIP6','Year'), 'ISIN', ('ISIN', 'Year')]
        :type id_cols: list of tuples
        :param data_source:
        :type data_source:
        :param data_loader: function that takes self.data_source as the arg, and usecols=self.id_cols as kwargs
        :type data_loader: Function
        :param data_loader_kwargs:
        :type data_loader_kwargs:
        """

        self.data_source = data_source
        self._df = None
        super().__init__(name)
        if id_cols is not None:
            self.id_cols = _id_cols_collection_to_flattened_list(id_cols)
            self.set_ids(id_cols)
        else:
            self.id_cols = []
            self.ids = []
        self.data_loader = _data_loader_or_pd_read_csv(data_loader)
        self.data_loader_kwargs = data_loader_kwargs

    @property
    def df(self):
        if self._df is None:
            if isinstance(self.data_source, pd.DataFrame):
                self._df = _dummy_load(self.data_source, self.id_cols)
            else:
                self._df = self.data_loader(self.data_source, usecols=self.id_cols, **self.data_loader_kwargs)
        return self._df

    def set_ids(self, id_cols):
        id_sets = _extract_id_sets(self.df, id_cols, name=self.name)
        self.ids = IDCollection(id_sets)

    def to_node(self, **node_kwargs):
        return Node(name=self.name, **node_kwargs)

    def __repr__(self):
        return f'<DataSource(name={self.name}, data_source={self.data_source})>'

class DataSubject:

    def __init__(self, *sources: [DataSource, 'DataCombination'], name=None):
        self.sources = sources
        self.name = name

    def __getitem__(self, item):
        return self.sources[item]

    def to_subgraph(self, options: PipelineOptions):
        nodes = self.to_nodes(options)

        return Subgraph(nodes, name=self.name, digraph=options.digraph, **options.graph_attr)

    def to_nodes(self, options: PipelineOptions):
        if options.data_sources:
            nodes = [source.to_node(**options.node_attr) for source in self.sources]
        else:
            node_kwargs = options.node_attr.copy()
            _add_no_node_line_to_node_kwargs(node_kwargs)
            nodes = [Node(name=self.name + 'DUMMY', label=None, **node_kwargs)]
        return nodes

    def __repr__(self):
        return f'<DataSubject(name={self.name}, sources={self.sources})>'

class DataCombination(DataItem):

    def __init__(self, data_source: DataSource, other_data_source: DataSource, name=None):
        self.data_source = data_source
        self.other_data_source = other_data_source
        super().__init__(name)
        self._ids = None

    def compare(self):
        return compare_id_collections(self.data_source.ids, self.other_data_source.ids, name=self.name)

    def to_node(self, plt=None, **node_kwargs):
        if plt is None:
            import matplotlib.pyplot as plt

        # TODO [#3]: finish compare ids data combination to node

        #### TEMP
        import pdb
        pdb.set_trace()
        #### END TEMP

        comparison = self.compare()
        # No ids, could not generate comparison, display normal node
        if not comparison:
            return Node(name=self.name, **node_kwargs)

        # Implicit else, has ids, able to generate comparison, now put graph in node
        graph = MatchComparisonBarGraph.from_id_comparison_collection(comparison, plt=plt)

        #### TEMP
        graph.draw(plt=plt)
        name_for_path = self.name.replace('/','-') + '.png'
        img_path = os.path.join(r'C:\Users\derobertisna.UFAD\Dropbox (Personal)\UF\Andy\ETF Project\Temp\idcomp', name_for_path)
        plt.savefig(img_path, bbox_inches='tight')

        # Set node style
        this_node_kwargs = node_kwargs.copy()
        _add_no_node_line_to_node_kwargs(this_node_kwargs)

        return Node(name=self.name, label=None, image=img_path, **this_node_kwargs)

        #### END TEMP

    @property
    def ids(self):
        if self._ids is None:
            # Find overlapping id sets
            id_sets = [id_set for id_set in self.data_source.ids if id_set in self.other_data_source.ids]
            self._ids = IDCollection(id_sets)
        return self._ids

    def __repr__(self):
        return f'<DataCombination(name={self.name}, data_source={self.data_source.name}, merge_source={self.other_data_source.name})>'



def _data_loader_or_pd_read_csv(data_loader=None):
    if data_loader is not None:
        return data_loader
    else:
        return pd.read_csv

def _dummy_load(df, usecols=None):
    if usecols is None:
        return df
    return df[usecols]

def _extract_id_sets(df, id_cols, name=None):

    collection_items = []
    for ids in id_cols:
        id_set_name = _ids_str_or_tuple_to_id_name(ids)
        unique = df[ids].drop_duplicates().dropna()
        # 'itertuples' for dfs, 'tolist' for series
        _extract_from_df_or_series_attr, kwargs = _get_func_attr_and_kwargs_to_reduce_from_df_or_series(unique)
        reduce_func = getattr(unique, _extract_from_df_or_series_attr)
        id_rows = reduce_func(**kwargs)
        id_set = IDSet(id_rows, name=id_set_name)
        collection_items.append(id_set)

    return IDCollection(collection_items, name=name)

def _id_cols_collection_to_flattened_list(id_cols):
    out_list = []
    for col_or_tuple in id_cols:
        if isinstance(col_or_tuple, (tuple, list)):
            [out_list.append(item) for item in col_or_tuple]
        else:
            out_list.append(col_or_tuple)
    return out_list

def _get_func_attr_and_kwargs_to_reduce_from_df_or_series(df_or_series):
    if isinstance(df_or_series, pd.DataFrame):
        return 'itertuples', {'index': False, 'name':'ids'}
    elif isinstance(df_or_series, pd.Series):
        return 'tolist', {}
    else:
        raise ValueError(f'Must pass series or dataframe, got type {type(df_or_series)}')


def _add_no_node_line_to_node_kwargs(node_kwargs):
    """
    NOTE: inplace
    """
    # Set node style
    no_line_str = 'setlinewidth(0)'
    if 'style' in node_kwargs:
        style = node_kwargs['style'] + ',' + no_line_str
    else:
        style = no_line_str
    node_kwargs['style'] = style


import math
from typing import Optional, Dict, List

import pyexlatex as pl
import pandas as pd
import matplotlib.pyplot as plt

PLOTS_PER_FIG = 8


def variables_drop_panel(summ_df: pd.DataFrame, group_name: str, id_var: Optional[str] = None) -> pl.Panel:
    blank_label = pl.Label('', span=2 if id_var is not None else 1)
    real_label = pl.Label('Drop Statistics', span=5)
    labels = pl.LabelCollection([blank_label, real_label], underline='3-6')
    dt = pl.DataTable.from_df(summ_df, include_index=True, extra_header=labels)
    panel = pl.Panel.from_data_tables([dt], name=group_name)
    return panel


def variables_latex_figures(figs: Dict[str, plt.Figure], group_name: str,
                             sources_outfolder: str = '.') -> List[pl.Figure]:
    num_figures = math.ceil(len(figs) / PLOTS_PER_FIG)
    fig_tuples = list(figs.items())

    bot = 0
    top = PLOTS_PER_FIG
    out_figs = []
    for i in range(num_figures):
        if num_figures == 1:
            name = group_name
        else:
            name = f'{group_name} ({i + 1}/{num_figures})'
        selected_figs = dict(fig_tuples[bot:top])
        pl_fig = pl.Figure.from_dict_of_names_and_plt_figures(selected_figs, sources_outfolder, figure_name=name)
        out_figs.append(pl_fig)
        bot += PLOTS_PER_FIG
        top += PLOTS_PER_FIG

    return out_figs




class SEMException(Exception):
    pass


class SampleException(SEMException):
    pass


class ModelException(SEMException):
    pass


class SampleCovMatrixNotPositiveDefiniteException(SampleException):
    pass


class ModelCovMatrixNotPositiveDefiniteException(ModelException):
    pass



import pyexlatex.table as lt

import pandas as pd

from datacode.summarize import format_numbers_to_decimal_places
from datacode.typing import DfDict, Document
from datacode.typing import DocumentOrTables, DocumentOrTablesOrNone


def outlier_by_column_summary(bad_df_dict: DfDict, selected_orig_df_dict: DfDict,
                              keep_num_rows: int =40, output: bool =False,
                              outdir: str = None, as_document=True, author: str=None) -> DocumentOrTables:

    all_tables = []
    for col in bad_df_dict:
        all_tables.append(
            outlier_summary_for_col(
                bad_df_dict,
                selected_orig_df_dict,
                col,
                keep_num_rows=keep_num_rows,
                output=False,
                as_document=False
            )
        )


    all_tables = [table for table in all_tables if table is not None]

    full_title = 'Outlier Summary'

    document = Document.from_ambiguous_collection(
        all_tables,
        title=full_title,
        author=author
    )

    if output:
        assert outdir is not None
        document.to_pdf_and_move(
            outdir,
            outname=full_title,
            as_document=True
        )

    if as_document:
        return document
    else:
        return all_tables


def outlier_summary_for_col(bad_df_dict: DfDict, selected_orig_df_dict: DfDict,
                            col: str, keep_num_rows: int =40, output: bool =False,
                            outdir: str = None, as_document=True, author: str=None) -> DocumentOrTablesOrNone:

    bad_df = bad_df_dict[col]
    selected_orig_df = selected_orig_df_dict[col]

    if len(bad_df) == 0:
        print(f'No outliers for {col}. Will not add tables.')
        return None

    bad_table = _firm_list_table_from_df(
        bad_df,
        col,
        keep_num_rows=keep_num_rows,
        caption=f'Largest Outliers for {col}',
        below_text=f'''This table shows the largest outliers for {col}.''',
        output=False
    )

    selected_df_tables = []
    processed_rows = 0
    while processed_rows < len(selected_orig_df):
        selected_df_table = _firm_list_table_from_df(
            selected_orig_df.iloc[processed_rows:processed_rows + keep_num_rows],
            col,
            keep_num_rows=keep_num_rows,
            caption=f'Outlier Firm Series for {col}',
            below_text=f'''This table shows observations leading up to, including, and after outliers for {col}.''',
            output=False
        )
        selected_df_tables.append(selected_df_table)
        processed_rows += keep_num_rows

    full_title = 'Outlier Summary for {col}'

    document = Document.from_ambiguous_collection(
        [bad_table] + selected_df_tables,
        title=full_title,
        author=author
    )

    if output:
        assert outdir is not None
        document.to_pdf_and_move(
            outdir,
            outname=full_title,
            as_document=True
        )

    if as_document:
        return document
    else:
        return [bad_table] + selected_df_tables





def _firm_list_table_from_df(df: pd.DataFrame, col: str,
                             keep_num_rows: int =40, caption: str =None,
                             below_text: str =None, output: bool =False,
                             outdir: str =None) -> lt.Table:

    if caption is None:
        caption =f'Largest Outliers for {col}'

    if below_text is None:
        below_text = f'''
This table shows the largest outliers for {col}.
'''

    formatted_df = df.iloc[:keep_num_rows].applymap(format_numbers_to_decimal_places)

    align_str = 'll' + 'c' * (len(formatted_df.columns) - 2)

    table = lt.Table.from_list_of_lists_of_dfs(
        [[formatted_df]],
        caption=caption,
        below_text=below_text,
        align=align_str,
        landscape=True
    )

    if output:
        assert outdir is not None
        table.to_pdf_and_move(
            outdir,
            outname=caption
        )

    return table



from typing import List, Optional
import pandas as pd
import numpy as np


def merge_treated_and_control_calculate_comparison_score(treated_entity_df: pd.DataFrame,
                                                         control_entity_df: pd.DataFrame,
                                                         prob_treated_var: str,
                                                         time_var: str = 'Date',
                                                         minimum_length: Optional[int] = None) -> float:
    compare_arrays = comparison_arrays_from_treated_and_control_df(
        treated_entity_df,
        control_entity_df,
        prob_treated_var,
        time_var=time_var,
        minimum_length=minimum_length
    )
    return comparison_score(compare_arrays)


def comparison_arrays_from_treated_and_control_df(treated_entity_df: pd.DataFrame,
                                                  control_entity_df: pd.DataFrame,
                                                  prob_treated_var: str,
                                                  time_var: str = 'Date',
                                                  minimum_length: Optional[int] = None,
                                                  ) -> List[np.array]:
    merged = treated_entity_df.merge(control_entity_df, how='inner', on=[time_var])

    if minimum_length is not None:
        df_len = len(merged)
        if df_len < minimum_length:
            raise NotEnoughTimeObservationsException(f'only had {df_len} obs while {minimum_length} were required')

    compare_arrays = [
        merged[prob_treated_var + '_x'].values,
        merged[prob_treated_var + '_y'].values
    ]

    return compare_arrays


def comparison_score(compare_arrays: List[np.array]) -> float:
    _validate_compare_arrays(compare_arrays)
    squared_diffs = (compare_arrays[0] - compare_arrays[1]) ** 2
    return sum(squared_diffs) / len(squared_diffs)


def _validate_compare_arrays(compare_arrays: List[np.array]):
    if len(compare_arrays) != 2:
        raise ValueError('must pass two compare arrays')
    if len(compare_arrays[0]) != len(compare_arrays[1]):
        raise ValueError('passed arrays must be of the same size.')
    if len(compare_arrays[0]) == 0:
        raise EmptyArraysException('cannot produce comparison score from empty arrays')


class NotEnoughTimeObservationsException(Exception):
    pass


class EmptyArraysException(Exception):
    pass


import pandas as pd

import pd_utils
from pd_utils.pdutils import _to_list_if_str
from datacode.ff.fftypes import ListOrStr

def long_averages_to_wide_averages(df: pd.DataFrame, datevar: str='Date', byvars: ListOrStr=None,
                                   retvar: str='RET',
                                   dual_portvar: str='Market Equity Portfolio/B/M Portfolio') -> pd.DataFrame:

    byvars = _to_list_if_str(byvars)

    if byvars is not None:
        all_byvars = byvars + [datevar]
    else:
        all_byvars = [datevar]

    return pd_utils.long_to_wide(
        df,
        groupvars=all_byvars,
        values=retvar,
        colindex=dual_portvar,
        colindex_only=True
    )


from pandas.testing import assert_frame_equal

from datacode import DataSource, Column
from tests.pipeline.base import PipelineTest


class TestDataTransformationPipeline(PipelineTest):

    def test_create_and_run_transformation_pipeline_from_func(self):
        dtp = self.create_transformation_pipeline()
        dtp.execute()

        assert_frame_equal(dtp.df, self.expect_func_df)

    def test_create_and_run_transformation_pipeline_from_transform(self):
        dtp = self.create_transformation_pipeline(func=self.source_transform)
        dtp.execute()

        assert_frame_equal(dtp.df, self.expect_loaded_df_with_transform)

    def test_auto_run_pipeline_by_load_source_with_no_location(self):
        dtp = self.create_transformation_pipeline()

        ds = DataSource(pipeline=dtp, location=self.csv_path_output)
        df = ds.df
        assert_frame_equal(df, self.expect_func_df)

    def test_auto_run_pipeline_by_load_source_with_no_location_and_shared_columns(self):
        self.create_csv()
        all_cols = self.create_columns()

        def transform_func(source: DataSource) -> DataSource:
            new_ds = DataSource(df=source.df, columns=all_cols)
            return new_ds

        dtp = self.create_transformation_pipeline(func=transform_func)

        ds = DataSource(pipeline=dtp, location=self.csv_path_output, columns=all_cols)
        df = ds.df
        assert_frame_equal(df, self.expect_loaded_df_rename_only)

    def test_create_nested_transformation_pipeline(self):
        dtp = self.create_transformation_pipeline()

        dtp2 = self.create_transformation_pipeline(source=dtp)
        dtp2.execute()
        assert_frame_equal(dtp2.df, self.expect_df_double_source_transform)

    def test_create_nested_generation_pipeline(self):
        dgp = self.create_generator_pipeline()

        dtp = self.create_transformation_pipeline(source=dgp)
        dtp.execute()

        assert_frame_equal(dtp.df, self.expect_generated_transformed)

    def test_create_nested_merge_pipeline(self):
        dmp = self.create_merge_pipeline()

        dtp = self.create_transformation_pipeline(source=dmp)
        dtp.execute()

        assert_frame_equal(dtp.df, self.expect_merged_1_2_both_transformed)

    def test_original_variables_not_affected_by_transform(self):
        self.create_csv()
        a, b, c = self.create_variables()
        ac = Column(a, 'a')
        bc = Column(b, 'b')
        cc = Column(c, 'c')
        all_cols = [ac, bc, cc]
        source = self.create_source(df=None, columns=all_cols)
        dtp = self.create_transformation_pipeline(source=source)
        dtp.execute()

        assert not a.applied_transforms
        assert not b.applied_transforms
        assert not c.applied_transforms

    def test_transform_on_source_with_normal_and_transformed_of_same_variable(self):
        self.create_csv()
        a, b, c = self.create_variables(transform_data='cell', apply_transforms=False)
        ac = Column(a, 'a')
        bc = Column(b, 'b')
        cc = Column(c, 'c')
        all_cols = [ac, bc, cc]
        include_vars = [
            a,
            a.add_one_cell(),
            b,
            c
        ]
        source = self.create_source(df=None, columns=all_cols, load_variables=include_vars)
        dtp = self.create_transformation_pipeline(source=source)
        dtp.execute()

        assert_frame_equal(dtp.df, self.expect_func_df_with_a_and_a_transformed)



from datacode.ff.fftypes import (
    StrOrInt,
    DictofStrsandStrLists,
    GroupvarNgroupsDict,
    TwoStrTupleList,
    StrBoolDict,
    Tuple
)

def parse_model(model_selector: StrOrInt) -> StrOrInt:

    if isinstance(model_selector, int):
        if model_selector in (3, 5):
            return model_selector
        else:
            raise ValueError(f'expected model 3 or 5 factor. got {model_selector} factor')

    model_selector = model_selector.strip().lower()

    if model_selector in ('3', '5'):
        return int(model_selector)

    if _is_custom_str(model_selector):
        return 'custom'

    raise ValueError(f'could not parse model selector. pass 3, 5, or custom. got {model_selector}')

def _is_custom_str(model_str: str):
    return model_str in ('c', 'custom', 'cust')

def _validate_model(model_selector: StrOrInt,
                    custom_labels: DictofStrsandStrLists = None,
                    custom_groupvar_ngroups_dict: GroupvarNgroupsDict = None,
                    custom_pairings: TwoStrTupleList = None,
                    custom_low_minus_high_dict: StrBoolDict = None) -> None:
    model = parse_model(model_selector)

    # base model validation is done in parse_model, this is for custom model validation
    if model != 'custom':
        return

    for item in [custom_labels, custom_groupvar_ngroups_dict, custom_pairings, custom_low_minus_high_dict]:
        if item is None:
            raise ValueError('got custom FF model type. must pass all of custom_labels, '
                             'custom_groupvar_ngroups_dict, custom_pairings, '
                             'and custom_low_minus_high_dict.')




import contextlib
import sys


class DummyFile:
    def write(self, x): pass


@contextlib.contextmanager
def no_stdout():
    save_stdout = sys.stdout
    sys.stdout = DummyFile()
    yield
    sys.stdout = save_stdout



from typing import Sequence, Dict, Optional

import pandas as pd
import pyexlatex as pl

from datacode.models.variables import Variable
from datacode.summarize.coverage.counts import variables_drop_count_df_and_plt_figs
from datacode.summarize.coverage.to_tex import variables_drop_panel, variables_latex_figures


def variables_drop_obs_doc(df: pd.DataFrame, var_groups: Dict[str, Sequence[Variable]],
                           id_var: Optional[str] = None, sources_outfolder: str = '.') -> pl.Document:
    """
    Generates a summary document of coverage of variables

    Produce a summary document of first a table which shows for each variable,
    the count of non-missing observations, and the count of coins which have
    at least one non-missing observation. Then it shows statistics on what
    would happen to the observations if the variable was dropped from the sample.
    These numbers are calculated by looking at every combination of the
    variables and how much the observation count would increase if this
    variable was excluded from the analysis. Summary statistics are shown for that.
    Then the following pages in the document are histograms of the same analysis,
    showing the distribution of observation counts that could be regained if we
    excluded that variable across the combinations of the other variables.

    :param df:
    :param var_groups: keys are names of variable groups, values are variables in the group
    :param id_var:
    :param sources_outfolder:
    :return:
    """
    panels = []
    all_figures = []
    for group_name, selected_vars in var_groups.items():
        var_names = [var.name for var in selected_vars]
        summ_df, figs = variables_drop_count_df_and_plt_figs(df, var_names, id_var)
        panel = variables_drop_panel(summ_df, group_name, id_var)
        panels.append(panel)
        figures = variables_latex_figures(figs, group_name, sources_outfolder)
        all_figures.extend(figures)
    tab = pl.Table.from_panel_list(panels, caption='Variable Drop Analysis')
    doc = pl.Document([tab] + all_figures)
    return doc



from typing import Callable, Optional, Any, Union, Dict

from datacode.models.transform.source import SourceTransform
from datacode.models.pipeline.operations.operation import DataOperation, OperationOptions
from datacode.models.source import DataSource
from datacode.models.types import DataSourcesOrPipelines


class TransformOperation(DataOperation):
    """
    Data operation that takes one DataSource as an input and outputs a DataSource
    """
    options: 'TransformOptions'
    result: 'DataSource'

    def __init__(self, data_sources: DataSourcesOrPipelines, options: 'TransformOptions', **result_kwargs):
        super().__init__(
            data_sources,
            options,
            **result_kwargs
        )

    @property
    def data_source(self) -> DataSource:
        return self.data_sources[0]

    def _execute(self):
        ds = self.options.transform.apply(self.data_source, preserve_original=self.options.preserve_original)
        self.result.update_from_source(ds, exclude_attrs=('location', 'data_outputter_kwargs'))
        return self.result

    def summary(self, *summary_args, summary_method: str=None, summary_function: Callable=None,
                             summary_attr: str=None, **summary_method_kwargs):
        # TODO [#53]: better summary for DataTransformationPipeline
        print(f'Calls transform {self.options.transform} on existing '
              f'data source {self.data_source}')

    def describe(self):
        return self.summary()

    def __repr__(self):
        return f'<TransformOperation(data_source={self.data_source}, options={self.options})>'

    def _validate(self):
        self._validate_data_sources_and_options()

    def _validate_data_sources_and_options(self):
        if len(self.data_sources) > 1:
            raise ValueError(f'got more than one data source got DataTransformPipeline, not clear which to use. '
                             f'Data sources: {self.data_sources}')


class TransformOptions(OperationOptions):
    """
    Class for options passed to AnalysisOperations
    """
    op_class = TransformOperation
    repr_cols = [
        'func',
        'out_path',
        'allow_modifying_result',
        'preserve_original',
    ]

    def __init__(self, func: Union[Callable[[DataSource, Any], DataSource], SourceTransform],
                 preserve_original: bool = True, out_path: Optional[str] = None,
                 allow_modifying_result: bool = True, result_kwargs: Optional[Dict[str, Any]] = None,
                 transform_key: Optional[str] = None):
        """

        :param func:
        :param preserve_original:
        :param out_path:
        :param allow_modifying_result: When DataSources are directly linked to pipelines, loading
            source from pipeline can cause modifications in the pipeline's result source. Set to False
            to ensure it won't be modified (but uses more memory). Setting to False should only be needed
            if multiple sources load from the same pipeline in one session
        :param transform_key: Only used when passing callable instead of SourceTransform. Sets the key for the
            generated SourceTransform
        """
        if isinstance(func, SourceTransform):
            self.transform = func
        else:
            self.transform = SourceTransform.from_func(func, key=transform_key)

        self.func = func
        self.preserve_original = preserve_original
        self.out_path = out_path
        self.allow_modifying_result = allow_modifying_result
        self.result_kwargs = result_kwargs



import os
from copy import deepcopy
from typing import TYPE_CHECKING, Optional, Dict, Any

import pandas as pd
from mixins import ReprMixin

if TYPE_CHECKING:
    from datacode.models.source import DataSource
    from datacode.models.variables.variable import Variable


class DataOutputter(ReprMixin):
    repr_cols = ['to_location_kwargs', 'safe', 'preserve_original', 'save_calculated']

    def __init__(self, source: 'DataSource', to_location_kwargs: Optional[Dict[str, Any]] = None,
                 safe: bool = True, preserve_original: bool = True, save_calculated: bool = False):
        if to_location_kwargs is None:
            to_location_kwargs = {}

        self.source = source
        self.to_location_kwargs = to_location_kwargs
        self.safe = safe
        self.preserve_original = preserve_original
        self.save_calculated = save_calculated

        self._validate()

    def _validate(self):
        if self.source.location is None:
            raise ValueError(f'must pass location to DataSource {self.source} to output it')
        if self.safe:
            self._check_safety()

    def _check_safety(self):
        if not self.source.columns:
            raise DataOutputNotSafeException(f'DataSource {self.source} has no columns, '
                                             f'cannot determine how to output safely. Pass safe=False to bypass')

        if self._output_exists and len(self.source.load_variables) < len(self.source.columns):
            raise DataOutputNotSafeException(f'DataSource {self.source} has {len(self.source.columns)} columns '
                                             f'but only {len(self.source.load_variables)} loaded variables, so '
                                             f'would be deleting existing data. Pass safe=False to bypass')

        if len(self.source.load_variables) > len(self.source.columns):
            raise DataOutputNotSafeException(f'DataSource {self.source} has {len(self.source.columns)} columns '
                                             f'but {len(self.source.load_variables)} loaded variables, so '
                                             f'some variables would not be outputted. Pass safe=False to bypass ')

    @property
    def _output_exists(self) -> bool:
        return os.path.exists(self.source.location)

    def output(self):
        if self.preserve_original:
            df = deepcopy(self.source.df)
        else:
            df = self.source.df
        self.rename_columns(df)
        self.keep_necessary_cols(df)
        self.output_to_location(df)

    def rename_columns(self, df: pd.DataFrame):
        if not self.source.columns:
            return

        rename_dict = {}
        for variable in self.source.load_variables:
            if variable.calculation is not None and not self.save_calculated:
                continue
            col = self.source.col_for(variable)
            rename_dict[variable.name] = col.load_key

        df.rename(columns=rename_dict, inplace=True)
        df.rename_axis(index=rename_dict, inplace=True)

    def keep_necessary_cols(self, df: pd.DataFrame):
        if not self.source.columns:
            return

        keep_cols = [col for col in df.columns if col in self.source.col_load_keys]

        num_cols = len(keep_cols) + len(df.index.names)
        if self.safe and num_cols < len(self.source.columns):
            existing_cols = keep_cols + df.index.names
            missing_cols = [col for col in self.source.col_load_keys if col not in existing_cols]
            raise DataOutputNotSafeException(f'After keeping necessary columns, only outputting '
                                             f'{num_cols} '
                                             f'columns when DataSource '
                                             f'describes {len(self.source.columns)} columns. '
                                             f'Missing columns {missing_cols}. '
                                             f'Pass safe=False to bypass.')

        drop_cols = [col for col in df.columns if col not in keep_cols]
        df.drop(drop_cols, axis=1, inplace=True)

    def output_to_location(self, df: pd.DataFrame):
        if not self.source.index_cols:
            # No index columns, index should be useless autoincrementing range, don't output it
            output_index = False
        else:
            output_index = True
        config_dict = deepcopy(self.to_location_kwargs)
        config_dict.update(dict(index=output_index))

        # TODO [#60]: implement output to location types other than CSV
        df.to_csv(self.source.location, **config_dict)



class DataOutputNotSafeException(Exception):
    pass


from typing import List, Tuple, Optional
import pandas as pd
from regtools.interact import _interaction_tuple_to_var_name
from regtools import reg_for_each_yvar_and_produce_summary


def diff_reg_for_each_yvar_and_produce_summary(diff_df: pd.DataFrame, yvars: List[str], treated_var: str,
                                               treated_time_var: str = 'After', xvars: Optional[List[str]] = None):
    interaction_tuple = (treated_var, treated_time_var)
    interaction_varname = _interaction_tuple_to_var_name(interaction_tuple)

    all_xvars = [treated_var, treated_time_var]
    if xvars is not None:
        all_xvars.extend(xvars)

    reg_list, summ = reg_for_each_yvar_and_produce_summary(
        diff_df,
        yvars,
        all_xvars,
        [treated_var, treated_time_var, interaction_varname],
        interaction_tuples=[interaction_tuple]
    )

    t_name = '$t$-stat'
    _add_t_of_last_interaction_to_horizontal_summary(summ.tables[0], interaction_tuple, reg_list, t_name=t_name)
    column_order = [treated_var, treated_time_var, interaction_varname, t_name, 'Controls', 'Adj-R2', 'N']
    summ.tables[0] = summ.tables[0][column_order]

    return reg_list, summ

def _add_t_of_last_interaction_to_horizontal_summary(sdf: pd.DataFrame, interaction_tuple: Tuple[str, str],
                                                     reg_list, t_name: str = '$t$-stat') -> None:
    """
    note: inplace

    Args:
        sdf:
        interaction_tuple:
        reg_list:
        t_name:

    Returns:

    """

    interaction_varname = _interaction_tuple_to_var_name(interaction_tuple)
    interaction_stderr = pd.Series([r.bse[interaction_varname] for r in reg_list], index=sdf.index)
    sdf[t_name] = sdf[interaction_varname].apply(lambda x: float(x.strip('*'))) / interaction_stderr
    sdf[t_name] = sdf[t_name].apply(lambda x: f'{x:.2f}')


import pandas as pd
import os
import matplotlib.pyplot as plt
from pyexlatex.figure import Figure

def missing_pct_by_id_figure(df: pd.DataFrame, id_col: str, col_with_missings: str,
                             outname: str='Percentage Missing Obs by Firm',
                             outfolder: str='.') -> Figure:

    if outfolder is None:
        outfolder = '.'

    matplotlib_figure = _missing_pct_by_id_plot(
        df,
        id_col,
        col_with_missings
    )

    base_outpath = os.path.join(outfolder, outname)
    source_outpath = f'{base_outpath} source.pdf'
    matplotlib_figure.savefig(source_outpath)
    plt.clf()

    figure = Figure.from_dict_of_names_and_filepaths(
        {outname: source_outpath},
        figure_name=outname,
        landscape=True,
        position_str_name_dict={outname: r'[t]{0.9\linewidth}'}
    )

    figure.to_pdf_and_move(
        outfolder=outfolder,
        outname=outname
    )

    return figure

def _missing_pct_by_id_plot(df: pd.DataFrame, id_col: str, col_with_missings: str) -> plt.Figure:

    id_pct_missing_series = _missing_pct_by_id_series(
        df,
        id_col,
        col_with_missings
    )

    ax = id_pct_missing_series.plot(kind='hist', bins=50, figsize=(14, 6))
    fig = ax.get_figure()

    return fig

def _missing_pct_by_id_series(df: pd.DataFrame, id_col: str, col_with_missings: str) -> pd.Series:

    if '_ones' in df.columns:
        raise ValueError('will overwrite existing column _ones')

    df['_ones'] = 1  # temporary column never missing for counting
    id_pct_missing = 1 - (df.groupby(id_col)[col_with_missings].count() / df.groupby(id_col)['_ones'].count())
    df.drop('_ones', axis=1, inplace=True)

    return id_pct_missing


from typing import Type, Optional, Sequence, Union, List

from mixins import ReprMixin
from mixins.attrequals import EqOnAttrsMixin, EqHashMixin

from datacode.models.transform.transform import Transform


class DataType(EqHashMixin, EqOnAttrsMixin, ReprMixin):
    _recursive_hash_convert = True
    names: Sequence[str] = tuple()
    name_roots: Sequence[str] = tuple()
    equal_attrs : List[str] = [
        'py_class',
        'pd_class',
        'categorical',
        'ordered',
        'names',
    ]
    repr_cols = ['categorical', 'ordered']

    def __init__(self, py_class: Type, pd_class: Optional[Type] = None,
                 transforms: Optional[Sequence[Transform]] = None, is_numeric: bool = False,
                 categorical: bool = False,
                 ordered: bool = False):
        self.names = [name.lower() for name in self.names]

        if transforms is None:
            transforms = []
            
        if pd_class is None:
            pd_class = py_class

        self.py_class = py_class
        self.pd_class = pd_class
        self.transforms = transforms
        self.categorical = categorical
        self.ordered = ordered
        self.is_numeric = is_numeric

    @classmethod
    def from_str(cls, dtype: str, categorical: bool = False, ordered: bool = False):
        # TODO [#22]: eliminate repeated from_str methods in dtypes
        #
        # Currently the same from_str method is in the subclasses because they have a different __init__
        # Only int and float have different from_str methods, and both of those are the same. Create
        # mixin or intermediate classes to eliminate repeated code.
        raise NotImplementedError('must implement from_str in subclass of DataType')

    @property
    def read_file_arg(self) -> Union[Type, str]:
        """
        The argument which should be passed to read_csv and other file reading
        methods in Pandas to ensure the correct data type
        """
        if self.categorical:
            return 'category'
        return self.pd_class

    @property
    def index_arg(self) -> Union[Type, str]:
        """
        The argument which should be passed to index.astype
        :return:
        """
        return self.read_file_arg



from typing import Union, Tuple, Iterable, List

PortfolioName = Union[str, int]
PortfolioPair = Tuple[PortfolioName, PortfolioName]
PortfolioPairs = Iterable[PortfolioPair]

StrList = List[str]
StrListList = List[StrList]


import pandas as pd
from functools import partial
from datacode.psm.names import predict_treated_varname, accurate_prediction_name, accuracy_pct
from datacode.summarize import format_numbers_to_decimal_places

def summarize_predictions(predict_df: pd.DataFrame, treated_var: str, treated_name: str = 'Treated',
                          control_name: str = 'Control') -> pd.DataFrame:
    predict_treated_var = predict_treated_varname(treated_var)

    treated_series = predict_df[[treated_var, predict_treated_var]].sum().astype(int)
    control_series = (len(predict_df) - treated_series).astype(int)
    correct_predictions_series = predict_df.groupby(treated_var)[accurate_prediction_name].sum().astype(int)
    count_series = pd.Series([control_series[treated_var], treated_series[treated_var]])

    predict_summ = pd.concat([control_series, treated_series], axis=1)
    predict_summ = pd.concat([
        predict_summ,  # num obs and predicted obs
        pd.DataFrame(correct_predictions_series).T,  # num accurate
        pd.DataFrame(correct_predictions_series / count_series).T  # pct accurate
    ])
    predict_summ.columns = [control_name, treated_name]
    predict_summ.index = ['# Obs', '# Predicted', '# Accurate', accuracy_pct]
    predict_summ = _format_predictions(predict_summ)

    return predict_summ[[treated_name, control_name]]

def _format_predictions(df: pd.DataFrame, decimals: int = 2) -> pd.DataFrame:
    df.loc[accuracy_pct] = df.loc[accuracy_pct] * 100
    num_formatter = partial(format_numbers_to_decimal_places, decimals=decimals, coerce_ints=True)
    df = df.applymap(format_numbers_to_decimal_places)

    return df


import pandas as pd
import pyexlatex.table as lt

def periods_data_table(df: pd.DataFrame, id_col: str, datevar: str,
                       period_display_name: str='Period') -> lt.DataTable:

    period_counts_df = _period_counts_df(
        df,
        id_col,
        datevar
    )

    period_counts_dt = lt.DataTable.from_df(
        period_counts_df,
        extra_header=f'{id_col}-{period_display_name}s'
    )

    return period_counts_dt

def _period_counts_df(df: pd.DataFrame, id_col: str, datevar: str) -> pd.DataFrame:
    date_counts = df.dropna().groupby(id_col)[datevar].count()

    period_counts_df = pd.DataFrame(
        data=[
            (
                date_counts.mean(),
                date_counts.median(),
                date_counts.max()
            )
        ],
        columns=['Mean', 'Median', 'Max']
    )

    period_counts_df = period_counts_df.applymap(lambda x: f'{x:.0f}')

    return period_counts_df


import datetime
from typing import Union, Type

import numpy as np
import pandas as pd

from datacode.models.dtypes.base import DataType


class DateType(DataType):
    names = ('date',)

    def __init__(self, categorical: bool = False, ordered: bool = False):
        super().__init__(
            datetime.date,
            pd_class=np.datetime64,
            categorical=categorical,
            ordered=ordered,
        )

    @classmethod
    def from_str(cls, dtype: str, categorical: bool = False, ordered: bool = False):
        dtype = dtype.lower()
        if dtype not in cls.names:
            raise ValueError(f'Dtype {dtype} does not match valid names for {cls.__name__}: {cls.names}')
        return cls(
            categorical=categorical,
            ordered=ordered
        )

    @property
    def index_arg(self) -> Union[Type, str]:
        return pd.DatetimeTZDtype



from typing import Sequence, Callable, Optional, Any, Dict

from datacode.models.analysis import AnalysisResult
from datacode.models.pipeline.operations.operation import DataOperation, OperationOptions
from datacode.models.source import DataSource
from datacode.models.types import DataSourceOrPipeline, DataSourcesOrPipelines


class AnalysisOperation(DataOperation):
    """
    Data operation that takes one DataSource as an input and does not output a DataSource
    """

    def __init__(self, data_sources: DataSourcesOrPipelines, options: 'AnalysisOptions', **result_kwargs):
        super().__init__(
            data_sources,
            options,
            **result_kwargs
        )

    @property
    def data_source(self) -> DataSource:
        return self.data_sources[0]

    def _execute(self):
        self.result.result = self.options.func(self.data_source, **self.options.func_kwargs)
        return self.result

    def summary(self, *summary_args, summary_method: str=None, summary_function: Callable=None,
                             summary_attr: str=None, **summary_method_kwargs):
        # TODO [#45]: better summary for DataAnalysisPipeline
        print(f'Runs func {self.options.func.__name__} with kwargs {self.options.func_kwargs} on {self.data_source}:')
        print(f'{self.data_source.describe()}')

    def describe(self):
        return self.summary()

    def __repr__(self):
        return f'<AnalysisOperation(data_source={self.data_source}, options={self.options})>'

    def _validate(self):
        self._validate_data_sources_and_options()

    def _validate_data_sources_and_options(self):
        if len(self.data_sources) > 1:
            raise ValueError(f'got more than one data source got DataAnalysisPipeline, not clear which to use. '
                             f'Data sources: {self.data_sources}')


def analysis_result_to_file(result: AnalysisResult, out_path: str):
    with open(out_path, 'w') as f:
        f.write(str(result.result))


class AnalysisOptions(OperationOptions):
    """
    Class for options passed to AnalysisOperations
    """
    op_class = AnalysisOperation
    result_class = AnalysisResult

    def __init__(self, func: Callable[[DataSource, Any], Any],
                 output_func: Optional[Callable[['AnalysisResult', str], None]] = analysis_result_to_file,
                 out_path: Optional[str] = None, result_kwargs: Optional[Dict[str, Any]] = None,
                 **func_kwargs):
        self.func = func
        self.func_kwargs = func_kwargs
        self.analysis_output_func = output_func
        self.out_path = out_path
        self.result_kwargs = result_kwargs

    @property
    def can_output(self) -> bool:
        return self.out_path is not None and self.analysis_output_func is not None



import pandas as pd
import pyexlatex.table as lt

from datacode.typing import StrList, IntSequenceOrNone, FloatSequenceOrNone

from datacode.summarize.subset.missing.detail.byid import by_id_pct_long_df
from datacode.summarize.subset.missing.detail.textfuncs import missing_more_than_pct_str, num_or_pct

def missing_more_than_data_table(df: pd.DataFrame, id_col: str, col_with_missings: str,
                                 missing_tolerances: IntSequenceOrNone=(0, 1, 10, 50),
                                 missing_quantiles: FloatSequenceOrNone = None,
                                 missing_display_str: str='Missing',
                                 period_display_name: str='Period',
                                 pct_format_str: str = '.1f') -> lt.DataTable:
    missing_more_than_df = _missing_more_than_df(
        df,
        id_col,
        col_with_missings,
        missing_tolerances=missing_tolerances,
        missing_quantiles=missing_quantiles,
        missing_display_str=missing_display_str,
        pct_format_str=pct_format_str
    )

    missing_more_than_dt = lt.DataTable.from_df(
        missing_more_than_df,
        extra_header=f'% {missing_display_str} > {num_or_pct(missing_tolerances, missing_quantiles)} '
                     f'{period_display_name}s'
    )

    return missing_more_than_dt

def _missing_more_than_df(df: pd.DataFrame, id_col: str, col_with_missings: str,
                          missing_tolerances: IntSequenceOrNone=(0, 1, 10, 50),
                          missing_quantiles: FloatSequenceOrNone = None,
                          missing_display_str: str='Missing',
                          pct_format_str: str = '.1f'
                          ) -> pd.DataFrame:

    if '_ones' in df.columns:
        raise ValueError('must not have a column _ones, will be overwritten')
    df['_ones'] = 1

    if missing_quantiles is not None:
        missing_cuts = missing_quantiles
        name_arg_index = 1
        cut_name = 'missing_quantile'
    elif missing_tolerances is not None:
        missing_cuts = missing_tolerances
        name_arg_index = 0
        cut_name = 'missing_tolerance'

    else:
        raise ValueError('must pass missing_tolerances or missing_quantiles, got both None')

    missing_more_than_df = pd.DataFrame()
    name_args = [None, None, missing_display_str, id_col]
    for missing_cut in missing_cuts:
        name_args[name_arg_index] = missing_cut
        missing_more_than_name = missing_more_than_pct_str(*name_args)
        cut_kwargs = {cut_name: missing_cut}
        rename_dict = {missing_more_than_name: f'{missing_cut * 100:g}%'}
        temp_by_id_pct_df = by_id_pct_long_df(
            df,
            ['_ones'],
            id_col,
            col_with_missings,
            missing_display_str=missing_display_str,
            **cut_kwargs
        ).drop([f'{id_col} Count', '_ones'], axis=1).rename(columns=rename_dict)
        missing_more_than_df = pd.concat([missing_more_than_df, temp_by_id_pct_df], axis=1)
    missing_more_than_df = missing_more_than_df.applymap(lambda x: f'{x:{pct_format_str}}')

    df.drop('_ones', axis=1, inplace=True)

    return missing_more_than_df


from typing import Any, Optional

from mixins import ReprMixin


class AnalysisResult(ReprMixin):
    repr_cols = ['name', 'location', 'result']

    def __init__(self, result: Any = None, name: Optional[str] = None,
                 location: Optional[str] = None):
        self.result = result
        self.name = name
        self.location = location



import graphviz
import uuid

class Node:

    def __init__(self, name, label='name', **node_kwargs):
        if label == 'name':
            self.label = name
        elif label is None:
            self.label = ''
        else:
            self.label = label

        self.name = name
        self.id = str(uuid.uuid1())
        self.node_kwargs = node_kwargs

    def add_to_graph(self, graph):
        graph.node(self.id, label=self.label, **self.node_kwargs)

    def __repr__(self):
        return f'<Node(name={self.name})>'

class NodeCollection:

    def __init__(self, *nodes):
        self.nodes = nodes

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.nodes[item]
        elif isinstance(item, str):
            return [node for node in self.nodes if node.name in (item, item + 'DUMMY')][0]

    def __repr__(self):
        return f'<NodeCollection(nodes={self.nodes})>'


class Subgraph:

    def __init__(self, nodes: [Node], name=None, digraph=True, **graph_kwargs):
        self.nodes = NodeCollection(*nodes)
        self.name = name
        self.digraph = digraph
        self.graph_kwargs = graph_kwargs

    def add_to_graph(self, graph):
        full_name = 'cluster' + (self.name if self.name else '')
        Graph = _get_digraph_if_true_else_graph(self.digraph)
        subgraph = Graph(full_name, **self.graph_kwargs)
        subgraph.attr('graph', label=self.name)
        [node.add_to_graph(subgraph) for node in self.nodes]
        graph.subgraph(subgraph)

    def __repr__(self):
        return f'<Subgraph(name={self.name}, nodes={self.nodes})>'

class Edge:

    def __init__(self, start: Node, end: Node, for_subgraphs: [None, (Subgraph,)] = None, **edge_kwargs):
        self.start = start
        self.end = end
        if for_subgraphs is not None:
            edge_kwargs.update({'ltail': 'cluster' + for_subgraphs[0].name, 'lhead': 'cluster' + for_subgraphs[1].name})
        self.edge_kwargs = edge_kwargs

    def add_to_graph(self, graph):
        graph.edge(self.start.id, self.end.id, **self.edge_kwargs)

    def __repr__(self):
        return f'<Edge(start={self.start}, end={self.end}, edge_kwargs: {self.edge_kwargs})>'


def _get_digraph_if_true_else_graph(boolean):
    if boolean:
        return graphviz.Digraph
    else:
        return graphviz.Graph



import uuid
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, Optional, List, Tuple

import pandas as pd
from mixins import ReprMixin

from datacode.models.column.column import Column
from pd_utils.optimize.load import read_file

from datacode.models.dtypes.date_type import DateType
from datacode.models.dtypes.datetime_type import DatetimeType

if TYPE_CHECKING:
    from datacode.models.source import DataSource
    from datacode.models.variables.variable import Variable


class DataLoader(ReprMixin):
    repr_cols = ['read_file_kwargs']

    def __init__(self, source: 'DataSource', read_file_kwargs: Optional[Dict[str, Any]] = None,
                 optimize_size: bool = False):
        if read_file_kwargs is None:
            read_file_kwargs = {}
        self.source = source
        self.optimize_size = optimize_size
        self.read_file_kwargs = read_file_kwargs

        # First var in tup is original variable, second is variable which needs column created
        self._calculated_variables_that_need_duplication: Dict[str, Tuple[Variable, Variable]] = {}

    def load_from_location(self) -> pd.DataFrame:
        """
        Used when df does not already exist in the source, loads it from location

        :return:
        """
        self.pre_read()
        df = self.read_file_into_df()
        df = self.post_read(df)
        df = self.duplicate_columns_for_calculations_assign_series(df)
        self.rename_columns(df)
        df = self.post_rename(df)
        if self.optimize_size:
            df = self.optimize_df_size(df)
        self.set_df_index(df)
        df = self.apply_calculations_transforms_and_drops(df)

        return df

    def load_from_existing_source(self, other_source: 'DataSource', preserve_original: bool = False) -> pd.DataFrame:
        """
        Used when df already exists in another source, just selects load_variables and
        applies calculations and transformations

        :return:
        """
        if preserve_original:
            df = other_source.df.copy()
        else:
            df = other_source.df

        # TODO [#59]: when loading from existing source, handle when indices do not match
        #
        # Currently the code assumes the same index in the existing source and loaded source.
        # Need to add code to change the index. But if this was due to a desired aggregation,
        # how should the user select what aggregation would be applied?

        self.select_variables_in_existing_source(df)
        df = self.apply_calculations_transforms_and_drops(df)
        return df

    def apply_calculations_transforms_and_drops(self, df: pd.DataFrame):
        self.assign_series_to_columns(df)
        df = self.pre_calculate(df)
        df = self.try_to_calculate_variables(df)
        self.duplicate_calculated_columns_if_necessary(df)
        df = self.pre_transform(df)
        df = self.apply_transforms(df)
        df = self.post_transform(df)
        df = self.try_to_calculate_variables(df)
        self.assign_series_to_columns(df)
        self.drop_variables(df)
        df = self.post_load(df)
        return df

    def duplicate_calculated_columns_if_necessary(self, df: pd.DataFrame):
        for var_key, (orig_var, new_var) in self._calculated_variables_that_need_duplication.items():
            self.source._duplicate_column_for_calculation(
                df,
                orig_var=orig_var,
                new_var=new_var,
                pre_rename=False,
            )

    def select_variables_in_existing_source(self, df: pd.DataFrame):
        if not (self.source.load_variables and self.source.columns):
            return

        keep_names = []
        for var in self.source.load_variables:
            for col in self.source.columns:
                if col.variable.key == var.key:
                    # Got column matching the desired variable
                    keep_names.append(var.name)  # add the variable name in the dataset to keep_names
        drop_cols = [col for col in df.columns if col not in keep_names]
        df.drop(drop_cols, axis=1, inplace=True)

    def read_file_into_df(self) -> pd.DataFrame:
        if self.source.location is None:
            return pd.DataFrame()

        read_file_config = dict()

        # Set which columns to load
        if self.source.load_variables and self.source.columns:
            usecols = []
            for var in self.source.load_variables:
                for col in self.source.columns:
                    if col.variable.key == var.key:
                        # Got column matching the desired variable
                        usecols.append(col.load_key)  # add the original column name in the dataset to usecols
            read_file_config['usecols'] = usecols

        # Set the data types of the columns
        date_dtypes = []
        datetime_dtypes = {}  # pandas requires separate handling for datetime
        if self.source.columns:
            dtypes = {}
            for col in self.source.columns:
                if col.dtype is not None:
                    if isinstance(col.dtype, DatetimeType):
                        # Track datetime separately
                        datetime_dtypes[col.load_key] = col.dtype
                    elif isinstance(col.dtype, DateType):
                        datetime_dtypes[col.load_key] = col.dtype
                        date_dtypes.append(col.load_key)
                    else:
                        dtypes[col.load_key] = col.dtype.read_file_arg
            if dtypes:
                read_file_config['dtype'] = dtypes
            if datetime_dtypes:
                read_file_config['parse_dates'] = list(datetime_dtypes.keys())


        read_file_config.update(self.read_file_kwargs)

        df = read_file(self.source.location, **read_file_config)

        # Extra cleanup for dates
        for col_key, dtype in datetime_dtypes.items():
            if isinstance(dtype, DateType):
                # Remove timestamp portion from date types
                df[col_key] = df[col_key].dt.date
            elif isinstance(dtype, DatetimeType) and dtype.tz is not None:
                # Convert to time zone stored in dtype
                try:
                    df[col_key] = df[col_key].dt.tz_localize(dtype.tz)
                except TypeError as e:
                    if 'Already tz-aware, use tz_convert' in str(e):
                        df[col_key] = df[col_key].dt.tz_convert(dtype.tz)
                    else:
                        raise e

        return df

    def set_df_index(self, df: pd.DataFrame):
        if not self.source.index_vars:
            return

        df.set_index(self.source.index_var_names, inplace=True)

    def assign_series_to_columns(self, df: pd.DataFrame):
        if not self.source.columns:
            return
        for var in self.source.load_variables:
            if var.key not in self.source.col_var_keys:
                if var.calculation is None:
                    raise ValueError(f'passed variable {var} but not calculated and not '
                                     f'in columns {self.source.columns}')
                continue
            col = self.source.col_for(var)
            series = self.source.get_series_for(var=var, df=df)
            col.series = series

    def duplicate_columns_for_calculations_assign_series(self, df: pd.DataFrame) -> pd.DataFrame:
        # TODO [#39]: more efficient implementation of loading variables for calculations
        #
        # The `DataLoader` checks what variables are needed for calculations that are not
        # included in `load_variables`, and if it requires multiple transformations of
        # a variable, then it copies that series for as many transformations are needed.
        # It would be better to have an implementation that doesn't require carrying copies
        # through everything.

        for col in self.source._columns_for_calculate:
            # Extra column is already in source, but need to add to df
            self.source._create_series_in_df_for_calculation(df, col)

        if not self.source.load_variables:
            return df

        # Now need to see if we need multiple transformations of a variable, then also copy that column
        # so that it won't get used up for just one of the transformations/original variable
        unique_var_keys: Dict[str, Variable] = {}
        for var in self.source.load_variables:
            # If the variable is needed multiple times, but not because of calculations (multiple transforms)
            if var.key in unique_var_keys and var not in self.source._vars_for_calculate:
                if var.calculation is not None:
                    # This calculated variable needs to be duplicated, but it has not been calculated yet.
                    # Therefore add to a list of variables which need to be duplicated after calculation.
                    if var.key not in self._calculated_variables_that_need_duplication:
                        self._calculated_variables_that_need_duplication[var.key] = (unique_var_keys[var.key], var)
                    continue
                # Got a variable multiple times, duplicate the column
                # Use the original variable for duplication as the column will already exist for that variable
                self.source._duplicate_column_for_calculation(
                    df,
                    orig_var=unique_var_keys[var.key],
                    new_var=var,
                )
            else:
                # Not a repeated variable, just add to tracking dict
                unique_var_keys[var.key] = var

        # Reorder df to be the same order as passed load variables
        col_order: Dict[str, int] = {}
        for col in self.source.columns:
            if col.variable.key not in self.source.load_var_keys:
                # Must be an unloaded column, skip it
                continue
            if col.variable.calculation is not None:
                # Calculated variables won't be in the data yet, skip them
                continue
            order = self.source.load_var_keys.index(col.variable.key)
            col_order[col.load_key] = order
        order_tups = list(col_order.items())
        order_tups.sort(key=lambda x: x[1])
        col_keys = [key for key, order in order_tups]
        df = df[col_keys]

        return df

    def optimize_df_size(self, df: pd.DataFrame) -> pd.DataFrame:
        # TODO [#17]: implement df size optimization
        #
        # Needs to be after adding data types to variables. Then can use data types to optimize
        raise NotImplementedError('implement df size optimization')

    def rename_columns(self, df: pd.DataFrame):
        from datacode.models.source import NoColumnForVariableException
        if not self.source.columns:
            return

        rename_dict = {}
        for variable in self.source._orig_load_variables:
            if variable.key not in self.source.col_var_keys:
                if variable.calculation is None:
                    raise ValueError(f'passed variable {variable} but not calculated and not '
                                     f'in columns {self.source.columns}')
                continue
            col = self.source.col_for(variable)
            rename_dict[col.load_key] = variable.name
            col.variable = variable
        for variable in self.source._vars_for_calculate:
            try:
                col = self.source.col_for(variable, for_calculate_only=True)
                rename_dict[col.load_key] = variable.name
                col.variable = variable
            except NoColumnForVariableException:
                # Must be using a pre-existing column rather than a newly generated column, need to rename that instead
                col = self.source.col_for(variable, orig_only=True)
                rename_dict[col.load_key] = variable.name
                col.variable = variable
        df.rename(columns=rename_dict, inplace=True)

    def try_to_calculate_variables(self, df: pd.DataFrame):
        if not self.source.columns:
            return df

        # Create temporary source so that transform can have access to df and all columns with one object
        temp_source = deepcopy(self.source)
        temp_source.df = df
        temp_source.name = '_temp_source_for_transform'

        for variable in self.source.load_variables:
            if variable.key in self.source.col_var_keys:
                # Variable already exists in the data, either from original source or previously calculated
                continue

            if variable.calculation is None:
                raise ValueError(f'passed variable {variable} but not calculated and not '
                                 f'in columns {self.source.columns}')
            required_variables = variable.calculation.variables
            has_all_required_variables = True
            calc_with_cols = []
            for req_var in required_variables:
                if not has_all_required_variables:
                    break
                col = self.source.col_for(req_var)
                calc_with_cols.append(col)
                col_pre_applied_transform_keys = deepcopy(col.applied_transform_keys)
                for transform in req_var.applied_transforms:
                    # Need to make sure all the same transforms have been applied to
                    # the column before the calculation
                    if transform.key in col_pre_applied_transform_keys:
                        col_pre_applied_transform_keys.remove(transform.key)
                    else:
                        has_all_required_variables = False
                        break

            if has_all_required_variables:
                # Actually do calculation
                new_series = variable.calculation.func(calc_with_cols)
                new_series.name = variable.name
                # TODO [#34]: determine how to set index for columns from calculated variables
                new_col = Column(variable, dtype=str(new_series.dtype), series=new_series)
                temp_source.df[variable.name] = new_series
                self.source.columns.append(new_col)

        return temp_source.df

    def apply_transforms(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self.source.columns:
            return df

        # Create temporary source so that transform can have access to df and all columns with one object
        # TODO [#28]: don't copy df, use same df
        temp_source = deepcopy(self.source)
        temp_source.df = df
        temp_source.name = '_temp_source_for_transform'

        for var in self.source.load_variables:
            if not var.applied_transforms:
                continue
            if var.key not in self.source.col_var_keys:
                if var.calculation is None:
                    raise ValueError(f'passed variable {var} but not calculated and not '
                                     f'in columns {self.source.columns}')
                continue
            column = self.source.col_for(var)
            temp_source = _apply_transforms_to_var(var, column, temp_source)
        return temp_source.df

    def drop_variables(self, df: pd.DataFrame):
        if not self.source._vars_for_calculate:
            # Only need to drop if extra variables were loaded for calculations
            return

        drop_names = [var.name for var in self.source._vars_for_calculate]
        df.drop(drop_names, axis=1, inplace=True)

    def pre_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def post_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def pre_read(self):
        pass

    def post_read(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def post_rename(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def pre_calculate(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def post_load(self, df: pd.DataFrame) -> pd.DataFrame:
        return df


def _apply_transforms_to_var(var: 'Variable', column: Column, source: 'DataSource') -> 'DataSource':
    col_pre_applied_transform_keys = deepcopy(column.applied_transform_keys)
    for transform in var.applied_transforms:
        if transform.key in col_pre_applied_transform_keys:
            # Transformation was already applied in the saved data source, skip this transformation
            # remove from applied transformations, because same transformation may be applied multiple times.
            # If desired transformation happens twice, and it is only once in the source column, will still
            # need to apply it once
            col_pre_applied_transform_keys.remove(transform.key)
            continue
        source = transform._apply_transform_for_column_and_variable_to_source(source, column, var)
        column.applied_transform_keys.append(transform.key)
        column.variable = var  # overwrite untransformed variable with transformed variable
    return source



import pandas as pd

from pyexlatex import Document
from datacode.summarize.subset.outliers.select import outlier_summary_dicts
from datacode.summarize.subset.outliers.detail.totex import outlier_by_column_summary
from datacode.summarize.subset.outliers.summary.main import outlier_overview_summary_page_table
from datacode.summarize.subset.outliers.typing import (
    AssociatedColDict,
    MinMaxDict,
    BoolDict,
    StrList,
    FloatSequence
)
from datacode.typing import DocumentOrTables


def outlier_summary_tables(df: pd.DataFrame, associated_col_dict: AssociatedColDict,
                           min_max_dict: MinMaxDict,
                           ascending_sort_dict: BoolDict = None,
                           always_associated_cols: StrList = None, bad_column_name: str = '_bad_column',
                           num_firms: int = 3, firm_id_col: str = 'TICKER',
                           date_col: str = 'Date',
                           begin_datevar: str = 'Begin Date',
                           end_datevar: str = 'End Date',
                           expand_months: int = 3,
                           keep_num_rows: int = 40, output: bool = True,
                           outdir: str = None, as_document=True, author: str = None,
                           title: str=None,
                           overview_caption: str='Outlier Overview and Summary Statistics',
                           quants: FloatSequence=(0.01, 0.1, 0.5, 0.9, 0.99)
                           ) -> DocumentOrTables:

    bad_df_dict, selected_orig_df_dict, bad_df = outlier_summary_dicts(
        df,
        associated_col_dict,
        min_max_dict,
        ascending_sort_dict=ascending_sort_dict,
        always_associated_cols=always_associated_cols,
        num_firms=num_firms,
        firm_id_col=firm_id_col,
        date_col=date_col,
        expand_months=expand_months,
        begin_datevar=begin_datevar,
        end_datevar=end_datevar,
        bad_column_name=bad_column_name
    )

    overview_table = outlier_overview_summary_page_table(
        df,
        bad_df,
        min_max_dict,
        bad_column_name=bad_column_name,
        firm_id_col=firm_id_col,
        as_document=False,
        caption=overview_caption,
        quants=quants
    )

    detail_tables = outlier_by_column_summary(
        bad_df_dict,
        selected_orig_df_dict,
        keep_num_rows=keep_num_rows,
        output=output,
        outdir=outdir,
        as_document=False,
    )

    all_tables = [overview_table] + detail_tables

    if as_document:
        document = Document.from_ambiguous_collection(
            all_tables,
            author=author,
            title=title
        )
        if outdir:
            document.to_pdf_and_move(
                outdir,
                outname=title
            )
        return document

    return all_tables



import pandas as pd
import pyexlatex.table as lt

from datacode.typing import SimpleDfDict, StrOrNone, IntSequenceOrNone, IntOrNone, FloatSequenceOrNone, FloatOrNone
from datacode.summarize.subset.missing.detail.byid import by_id_pct_long_df
from datacode.summarize.subset.missing.detail.byobs import obs_pct_long_df
from datacode.summarize.subset.missing.detail.reformat.main import long_counts_to_formatted_wide_df_dict
from datacode.summarize.subset.missing.detail.totex import missing_detail_df_dict_to_table_and_output
from datacode.summarize.subset.missing.summary.tables.fullsample.main import missing_full_sample_summary_panel

def obs_and_id_count_and_missing_pct_table(df: pd.DataFrame, col_with_missings: str, id_col: str,
                                           row_byvar: str, col_byvar: str, datevar: str,
                                           missing_tolerance: IntOrNone=0, missing_quantile: FloatOrNone = None,
                                           summary_missing_tolerances: IntSequenceOrNone = (0, 5, 10),
                                           summary_missing_quantiles: FloatSequenceOrNone = None,
                                           sort_cols_as_numeric: bool = True, sort_rows_as_numeric: bool = True,
                                           sort_cols_as_portvar: bool = False, sort_rows_as_portvar: bool = False,
                                           count_format_str: str = '.0f', pct_format_str: str = '.1f',
                                           missing_display_str: str = 'Missing', period_display_name: str = 'Period',
                                           extra_caption: str = '', extra_below_text: str = '',
                                           table_align: str = None,
                                           outfolder: StrOrNone=None) -> lt.Table:

    df_dict = obs_and_id_count_and_missing_pct_df_dict(
        df,
        col_with_missings,
        id_col,
        row_byvar,
        col_byvar,
        missing_tolerance=missing_tolerance,
        missing_quantile=missing_quantile,
        sort_cols_as_numeric=sort_cols_as_numeric,
        sort_rows_as_numeric=sort_rows_as_numeric,
        sort_cols_as_portvar=sort_cols_as_portvar,
        sort_rows_as_portvar=sort_rows_as_portvar,
        count_format_str=count_format_str,
        pct_format_str=pct_format_str,
        missing_display_str=missing_display_str
    )

    summary_panel = missing_full_sample_summary_panel(
        df,
        id_col,
        col_with_missings,
        missing_tolerances=summary_missing_tolerances,
        missing_quantiles=summary_missing_quantiles,
        missing_display_str=missing_display_str,
        datevar=datevar,
        pct_format_str=pct_format_str,
        period_display_name=period_display_name
    )

    table = missing_detail_df_dict_to_table_and_output(
        df_dict,
        summary_panel,
        row_byvar,
        col_byvar,
        id_col,
        col_with_missings,
        missing_tolerance=missing_tolerance,
        missing_quantile=missing_quantile,
        summary_missing_tolerances=summary_missing_tolerances,
        summary_missing_quantiles=summary_missing_quantiles,
        missing_display_str=missing_display_str,
        period_display_name=period_display_name,
        extra_caption=extra_caption,
        extra_below_text=extra_below_text,
        align=table_align,
        outfolder=outfolder
    )

    return table


def obs_and_id_count_and_missing_pct_df_dict(df: pd.DataFrame, col_with_missings: str, id_col: str,
                                             row_byvar: str, col_byvar: str,
                                             missing_tolerance: int=0, missing_quantile: IntOrNone = None,
                                             sort_cols_as_numeric: bool = True, sort_rows_as_numeric: bool = True,
                                             sort_cols_as_portvar: bool=False, sort_rows_as_portvar: bool=False,
                                             count_format_str: str = '.0f', pct_format_str: str = '.1f',
                                             missing_display_str: str = 'Missing'
                                             ) -> SimpleDfDict:

    byvars = [row_byvar, col_byvar]

    common_args = (
        df,
        byvars,
        id_col,
        col_with_missings
    )

    obs_pct_df = obs_pct_long_df(
        *common_args,
        missing_display_str=missing_display_str
    )

    by_id_pct_df = by_id_pct_long_df(
        *common_args,
        missing_tolerance=missing_tolerance,
        missing_quantile=missing_quantile,
        missing_display_str=missing_display_str
    )

    format_strs = [count_format_str, pct_format_str]

    common_args = (
        row_byvar,
        col_byvar
    )

    common_kwargs = dict(
        sort_cols_as_numeric=sort_cols_as_numeric,
        sort_rows_as_numeric=sort_rows_as_numeric,
        sort_cols_as_portvar=sort_cols_as_portvar,
        sort_rows_as_portvar=sort_rows_as_portvar,
        format_strs=format_strs
    )

    df_dict = long_counts_to_formatted_wide_df_dict(
        obs_pct_df,
        *common_args,
        **common_kwargs
    )

    df_dict.update(
        long_counts_to_formatted_wide_df_dict(
            by_id_pct_df,
            *common_args,
            **common_kwargs
        )
    )

    return df_dict




from typing import Union, Optional

from datacode.models.dtypes.base import DataType
from datacode.models.dtypes.bool_type import BooleanType
from datacode.models.dtypes.date_type import DateType
from datacode.models.dtypes.datetime_type import DatetimeType
from datacode.models.dtypes.float_type import FloatType
from datacode.models.dtypes.int_type import IntType
from datacode.models.dtypes.manager import DataTypeManager
from datacode.models.dtypes.obj_type import ObjectType
from datacode.models.dtypes.period_type import PeriodType
from datacode.models.dtypes.str_type import StringType
from datacode.models.dtypes.timedelta_type import TimedeltaType


def convert_str_to_data_type_if_necessary(dtype: Union[str, DataType]) -> Optional[DataType]:
    if isinstance(dtype, DataType):
        return dtype

    if dtype is None:
        return None

    return convert_str_to_data_type(dtype)


def convert_str_to_data_type(dtype: str) -> DataType:
    manager = DataTypeManager([
        BooleanType,
        DateType,
        DatetimeType,
        FloatType,
        IntType,
        ObjectType,
        StringType,
        TimedeltaType,
        PeriodType,
    ])
    return manager.get_by_name(dtype)



import pandas as pd

from datacode.summarize import format_numbers_to_decimal_places
from datacode.summarize.subset.outliers.typing import MinMaxDict


def bad_df_summary_df(bad_df: pd.DataFrame, min_max_dict: MinMaxDict,
                      firm_id_col: str = 'TICKER',
                      bad_column_name: str = '_bad_column') -> pd.DataFrame:
    compare_col_df_list = []
    obs_count_name = 'Obs. Count'
    id_col_count_name = f'{firm_id_col} Count'

    # Count of observations
    obs_count_series = bad_df.groupby(bad_column_name)[firm_id_col].count()
    obs_count_series.name = obs_count_name
    compare_col_df_list.append(obs_count_series)

    # Count of unique ids in observations
    id_count_series = bad_df[[firm_id_col, bad_column_name]].drop_duplicates(
    ).groupby(bad_column_name)[firm_id_col].count()
    id_count_series.name = id_col_count_name
    compare_col_df_list.append(id_count_series)

    # Combine cutoffs with counts
    concat_items = compare_col_df_list + [_min_max_dict_to_df(min_max_dict)]
    summ_df = pd.concat(concat_items, axis=1)

    # Cutoffs df may contain rows which counts will not. This means there were no outliers
    # for that variable. Fill with zeroes
    summ_df.fillna(0, inplace=True)

    # Convert counts to int for correct formatting
    summ_df[obs_count_name] = summ_df[obs_count_name].astype(int)
    summ_df[id_col_count_name] = summ_df[id_col_count_name].astype(int)

    summ_df = summ_df.applymap(format_numbers_to_decimal_places)

    summ_df.index.name = 'Variable'

    return summ_df


def _min_max_dict_to_df(min_max_dict: MinMaxDict) -> pd.DataFrame:
    df = pd.DataFrame(min_max_dict).T
    df.columns = ['Low Cutoff', 'High Cutoff']

    return df


from typing import Tuple
import re

RESULT_PATTERN = re.compile(r'([^\*]+)(\**)')

def convert_to_stars(t_value: float) -> str:
    t = abs(t_value)
    if t < 1.645:
        return ''
    elif t < 1.96:
        return '*'
    elif t < 2.576:
        return '**'
    elif t > 2.576:
        return '***'
    else: # covers nan
        return ''


def parse_stars_value(value: str) -> Tuple[str, str]:
    match = RESULT_PATTERN.fullmatch(value)
    if not match:
        return '', ''

    result = match.group(1)
    stars = match.group(2)
    return result, stars



# For FF procedure, size is the main port

import pandas as pd

from datacode.ff.create.minus import (
    _get_minus_label,
    _get_low_minus_high_dict,
    _calculated_with_varname_to_orig_var
)
from datacode.ff.fftypes import (
    TwoStrTupleList,
    TwoStrListTuple,
    DictofStrsandStrLists,
    StrOrInt,
    StrBoolDict,
    Tuple,
    StrList
)

def combine_main_portfolios(df: pd.DataFrame, labels: DictofStrsandStrLists, pairings: TwoStrTupleList,
                            factor_model: StrOrInt=3,
                            size_var: str=None, value_var: str=None, profitability_var: str=None,
                            investment_var: str=None,
                            custom_low_minus_high_dict: StrBoolDict=None) -> None:
    """
    Note: inplace

    Args:
        df:
        labels:
        pairings:
        factor_model:
        byvars:
        size_var:
        value_var:
        profitability_var:
        investment_var:
        custom_low_minus_high_dict:

    Returns:

    """

    low_minus_high_dict = _get_low_minus_high_dict(
        factor_model=factor_model,
        size_var=size_var,
        value_var=value_var,
        profitability_var=profitability_var,
        investment_var=investment_var,
        custom_low_minus_high_dict=custom_low_minus_high_dict
    )

    # Split processing into repeated (need to be averaged) and unique (need to be renamed)
    repeated_main_ports, unique_main_ports = _get_repeated_and_unique_primary_portvars(pairings)
    _combine_repeated_ports(
        df=df,
        labels=labels,
        repeated_ports=repeated_main_ports,
        low_minus_high_dict=low_minus_high_dict
    )
    _rename_unique_ports(
        df=df,
        labels=labels,
        unique_ports=unique_main_ports,
        low_minus_high_dict=low_minus_high_dict
    )


def _combine_repeated_ports(df: pd.DataFrame, labels: DictofStrsandStrLists, repeated_ports: StrList,
                            low_minus_high_dict: StrBoolDict) -> None:
    for portvar in repeated_ports:
        low_minus_high = low_minus_high_dict[portvar]
        _combine_repeated_port(
            df=df,
            labels=labels,
            portvar=portvar,
            low_minus_high=low_minus_high
        )


def _combine_repeated_port(df: pd.DataFrame, labels: DictofStrsandStrLists, portvar: str='Market Equity', low_minus_high=False) -> None:
    """
    Note: inplace

    Args:
        df:
        labels:
        portvar:
        low_minus_high:

    Returns:

    """
    label, matching_ports = _get_label_and_matching_ports(
        df=df,
        labels=labels,
        portvar=portvar,
        low_minus_high=low_minus_high
    )

    df[label] = df[matching_ports].apply(lambda x: x.mean(), axis=1)
    df.drop(matching_ports, axis=1, inplace=True)


def _rename_unique_ports(df: pd.DataFrame, labels: DictofStrsandStrLists, unique_ports: StrList,
                            low_minus_high_dict: StrBoolDict) -> None:
    for portvar in unique_ports:
        low_minus_high = low_minus_high_dict[portvar]
        _rename_unique_port(
            df=df,
            labels=labels,
            portvar=portvar,
            low_minus_high=low_minus_high
        )


def _rename_unique_port(df: pd.DataFrame, labels: DictofStrsandStrLists, portvar: str='Market Equity', low_minus_high=False) -> None:
    """
    Note: inplace

    Args:
        df:
        labels:
        portvar:
        low_minus_high:

    Returns:

    """
    label, matching_ports = _get_label_and_matching_ports(
        df=df,
        labels=labels,
        portvar=portvar,
        low_minus_high=low_minus_high
    )

    if len(matching_ports) > 1:
        raise ValueError(f'got portvar {label} as unique, expected only one column match. '
                         f'got {len(matching_ports)} matches: {matching_ports}')

    df.rename(columns={matching_ports[0]: label}, inplace=True)


def _get_label_and_matching_ports(df: pd.DataFrame, labels: DictofStrsandStrLists, portvar: str='Market Equity', low_minus_high=False) -> Tuple[str, StrList]:
    label = _get_minus_label(
        labels=labels,
        portvar=portvar,
        low_minus_high=low_minus_high
    )

    matching_ports = [col for col in df.columns if _calculated_with_varname_to_orig_var(col) == label]

    return label, matching_ports

def _get_repeated_and_unique_primary_portvars(pairings: TwoStrTupleList) -> TwoStrListTuple:
    unique_portvars = []
    repeated_portvars = []
    for main_portvar, secondary_portvar in pairings:
        if main_portvar in repeated_portvars:
            # Third or later time seeing variable, don't need to do anything
            continue
        elif main_portvar in unique_portvars:
            # Second time seeing variable, remove from unique, add to repeated
            repeated_portvars.append(main_portvar)
            unique_portvars.remove(main_portvar)
        else:
            # First time seeing variable, set to unique
            unique_portvars.append(main_portvar)

    return repeated_portvars, unique_portvars



"""
Hooks to run functions during datacode operations globally

:Usage:

    >>> import datacode.hooks as dc_hooks
    ...
    >>> def log_begin_execute(pipeline: DataPipeline):
    ...     print(f"running {pipeline.name}")
    ...
    >>> dc_hooks.on_begin_execute_pipeline = log_begin_execute

:All the Hooks:

    Each hook may expect different arguments and have different return values.
    Here is the list of the hooks, along with the expected arguments and return values.

"""
from copy import deepcopy
from typing import TYPE_CHECKING, Tuple

import pandas as pd


if TYPE_CHECKING:
    from datacode.models.pipeline.base import DataPipeline
    from datacode.models.pipeline.operations.operation import DataOperation
    from datacode.models.source import DataSource
    from datacode.models.column.column import Column
    from datacode.models.variables.variable import Variable
    from datacode.models.transform.transform import Transform
    from datacode.models.transform.source import SourceTransform

_orig_locals = {
    key: value for key, value in locals().items() if not key.startswith("_")
}


def on_begin_execute_pipeline(pipeline: "DataPipeline") -> None:
    """
    Called at the beginning of :meth:`DataPipeline.execute`

    :param pipeline: The pipeline which is about to be executed
    :return: None
    """
    pass


def on_end_execute_pipeline(pipeline: "DataPipeline") -> None:
    """
    Called at the end of :meth:`DataPipeline.execute`

    :param pipeline: The pipeline which was just executed
    :return: None
    """
    pass


def on_begin_execute_operation(operation: "DataOperation") -> None:
    """
    Called at the beginning of :meth:`DataOperation.execute`

    :param operation: The operation which is about to be executed
    :return: None
    """
    pass


def on_end_execute_operation(operation: "DataOperation") -> None:
    """
    Called at the end of :meth:`DataOperation.execute`

    :param operation: The operation which was just executed
    :return: None
    """
    pass


def on_begin_load_source(source: "DataSource") -> None:
    """
    Called at the beginning of :meth:`DataSource._load`, which
    is usually called when loading :paramref:`.DataSource.df`
    from either location or pipeline.

    :param source: DataSource for which DataFrame is about to be loaded
    :return:
    """
    pass


def on_end_load_source(source: "DataSource", df: pd.DataFrame) -> pd.DataFrame:
    """
    Called at the end of :meth:`DataSource._load`, which
    is usually called when loading :paramref:`.DataSource.df`
    from either location or pipeline.

    The DataFrame returned from this function will be
    set as :paramref:`.DataSource.df`

    :param source: DataSource for which DataFrame was just loaded
    :param df:
    :return: DataFrame which will be set as :paramref:`.DataSource.df`
    """
    return df


def on_begin_apply_variable_transform(
    transform: "Transform", source: "DataSource", col: "Column", var: "Variable"
) -> Tuple["DataSource", "Column", "Variable"]:
    """
    Called at the beginning of :meth:`Transform._apply_transform_for_column_and_variable_to_source`,
    which is called while transforming variables.

    The DataSource, Column, and Variable returned from this function will be used
    to do the variable transform

    :param transform: Transform which is about to be applied
    :param source: DataSource in which the transform is about to be applied
    :param col: Column on which the transform is about to be applied
    :param var: Variable on which the transform is about to be applied
    :return: The items which will be used to transform the variable
    """
    return source, col, var


def on_end_apply_variable_transform(
    transform: "Transform", source: "DataSource", col: "Column", var: "Variable"
) -> "DataSource":
    """
    Called at the end of :meth:`Transform._apply_transform_for_column_and_variable_to_source`,
    which is called while transforming variables.

    The DataSource returned from this function will be used
    as the results of the variable transform

    :param transform: Transform which was just applied
    :param source: DataSource in which the transform was just applied
    :param col: Column on which the transform was just applied
    :param var: Variable on which the transform was just applied
    :return: The DataSource which will be used as the results
        of the variable transform
    """
    return source


def on_begin_apply_source_transform(transform: 'SourceTransform', source: 'DataSource') -> 'DataSource':
    """
    Called at the beginning of :meth:`SourceTransform.apply`, which
    is called in TransformPipeline

    The DataSource returned from this function will be used
    to do the source transform

    :param transform: SourceTransform which is about to be applied
    :param source: DataSource on which the SourceTransform will be applied
    :return: The DataSource on which to apply the source transform
    """
    return source


def on_end_apply_source_transform(transform: 'SourceTransform', source: 'DataSource') -> 'DataSource':
    """
    Called at the end of :meth:`SourceTransform.apply`, which
    is called in TransformPipeline

    The DataSource returned from this function will be used
    as the result of the source transform

    :param transform: SourceTransform which was just applied
    :param source: DataSource on which the SourceTransform was just applied
    :return: The DataSource which will become the result of
        the source transformation
    """
    return source


_new_locals = {key: value for key, value in locals().items() if not key.startswith("_")}
_hook_keys = [key for key in _new_locals if key not in _orig_locals]
_orig_hooks = deepcopy(
    {key: value for key, value in _new_locals.items() if key in _hook_keys}
)


def reset_hooks() -> None:
    """
    Go back to original dummy hooks, removes all user settings of hooks

    :return: None

    :Notes:

        This is the only function in the module which is not a hook itself.
        Instead it is a utility method meant to be called by the user.

    """
    for key, value in _orig_hooks.items():
        globals()[key] = value


__all__ = _hook_keys + ["reset_hooks"]



from typing import Sequence, Dict

from datacode.models.variables.variable import Variable


def model_str(structural_dict: Dict[Variable, Sequence[Variable]],
              measurement_dict: Dict[Variable, Sequence[Variable]],
              var_corr_groups: Sequence[Sequence[Variable]]) -> str:
    m_str = '# structural part\n'
    for y, x_vars in structural_dict.items():
        all_vars = [y, *x_vars]
        m_str += _vars_to_structural_str(all_vars)
        m_str += '\n'
    m_str += '\n# measurement part\n'
    for y, x_vars in measurement_dict.items():
        all_vars = [y, *x_vars]
        m_str += _vars_to_measurement_str(all_vars)
        m_str += '\n'
    m_str += '\n# correlations\n'
    for corr_group in var_corr_groups:
        m_str += _vars_to_correlated_str(corr_group)
        m_str += '\n'
    return m_str


def _vars_to_add_keys_str(var_seq: Sequence[Variable]) -> str:
    return ' + '.join([var.unique_key for var in var_seq])


def _vars_to_structural_str(var_seq: Sequence[Variable]) -> str:
    return _vars_to_eq_str(var_seq, '~')


def _vars_to_measurement_str(var_seq: Sequence[Variable]) -> str:
    return _vars_to_eq_str(var_seq, '=~')


def _vars_to_correlated_str(var_seq: Sequence[Variable]) -> str:
    return _vars_to_eq_str(var_seq, '~~')


def _vars_to_eq_str(var_seq: Sequence[Variable], eq_sym: str = '~') -> str:
    if len(var_seq) < 2:
        return ''

    return f'{var_seq[0].unique_key} {eq_sym} ' + _vars_to_add_keys_str(var_seq[1:])




import datetime
from copy import deepcopy
from typing import Sequence, List, Callable, Optional, Union

from graphviz import Digraph
from mixins import ReprMixin

from datacode.graph.base import GraphObject
from datacode.graph.edge import Edge
from datacode.graph.node import Node
from datacode.graph.subgraph import Subgraph
import datacode.hooks as hooks
from datacode.models.analysis import AnalysisResult
from datacode.models.pipeline.operations.operation import DataOperation, OperationOptions
from datacode.models.source import DataSource
from datacode.models.types import DataSourcesOrPipelines, DataSourceOrPipeline, ObjWithLastModified


class DataPipeline(ReprMixin):
    """
    Base class for data pipelines. Should not be used directly.
    """
    repr_cols = ['name', 'data_sources', 'operation_options']

    def __init__(self, data_sources: DataSourcesOrPipelines,
                 operation_options: Optional[Sequence[OperationOptions]],
                 name: Optional[str] = None):
        """

        :param data_sources:
        :param operation_options:
        :param name:
        """
        if operation_options is None:
            operation_options = []
        if data_sources is None:
            data_sources = []

        if not isinstance(data_sources, list):
            data_sources = list(data_sources)

        if not isinstance(operation_options, list):
            operation_options = list(operation_options)

        self.data_sources: List[DataSourceOrPipeline] = data_sources
        self.operation_options: List[OperationOptions] = operation_options
        self.name = name
        self.df = None
        self._operation_index = 0
        self.result = None
        self.primary_node = Node(self.name)

    def execute(self, output: bool = True):
        hooks.on_begin_execute_pipeline(self)
        while True:
            try:
                self.next_operation()
            except LastOperationFinishedException:
                break

        self.result = self.operations[-1].result

        if output:
            self.output()

        hooks.on_end_execute_pipeline(self)
        return self.df

    def next_operation(self):
        if self._operation_index == 0:
            self._set_df_from_first_operation()

        self._do_operation()

    def reset(self):
        self.df = None
        self._operation_index = 0
        self._set_operations()

    def _do_operation(self):
        try:
            operation = self.operations[self._operation_index]
            print(f'Now running operation {self._operation_index + 1}: {operation}')
        except IndexError:
            raise LastOperationFinishedException

        operation.execute()

        # Set current df to result of merge
        if isinstance(operation.result, DataSource):
            # Need to check as may be analysis result, in which case df should not be changed
            self.df = operation.result.df

        self._operation_index += 1

    def _set_operations(self):
        self._operations = self._create_operations(self.data_sources, self.operation_options)

    def _create_operations(self, data_sources: DataSourcesOrPipelines, options_list: List[OperationOptions]):
        if options_list[0].op_class.num_required_sources == 0:
            operations = [options_list[0].get_operation(options_list[0])]
        elif options_list[0].op_class.num_required_sources == 1:
            operations = _get_operations_for_single(data_sources[0], options_list[0])
        elif options_list[0].op_class.num_required_sources == 2:
            operations = _get_operations_for_pair(data_sources[0], data_sources[1], options_list[0])
        else:
            raise ValueError('DataPipeline cannot handle operations with more than two sources')

        if len(options_list) == 1:
            return operations

        for i, options in enumerate(options_list[1:]):
            if options.op_class.num_required_sources == 0:
                operations.append(options.get_operation(options))
            elif options.op_class.num_required_sources == 1:
                operations += _get_operations_for_single(operations[-1].result, options)
            elif options.op_class.num_required_sources == 2:
                operations += _get_operations_for_pair(operations[-1].result, data_sources[i + 2], options)
            else:
                raise ValueError('DataPipeline cannot handle operations with more than two sources')

        return operations

    def _set_df_from_first_operation(self):
        # Need to check as may be generation pipeline which would not have a df to start from
        if hasattr(self.operations[0], 'data_sources') and self.operations[0].data_sources:
            self.df = self.operations[0].data_sources[0].df

    @property
    def operations(self) -> List[DataOperation]:
        try:
            return self._operations
        except AttributeError:
            self._set_operations()

        return self._operations

    # Following properties exist to recreate operations if data sources or merge options are overridden
    # by user

    @property
    def data_sources(self) -> List[DataSourceOrPipeline]:
        return self._data_sources

    @data_sources.setter
    def data_sources(self, data_sources: List[DataSourceOrPipeline]):
        self._data_sources = data_sources
        # only set merges if previously set. otherwise no need to worry about updating cached result
        if hasattr(self, '_operations'):
            self._set_operations()

    @property
    def operation_options(self):
        return self._operations_options

    @operation_options.setter
    def operation_options(self, options: List[OperationOptions]):
        self._operations_options = options
        # only set merges if previously set. otherwise no need to worry about updating cached result
        if hasattr(self, '_operations'):
            self._set_operations()

    def output(self):
        if self.result is None:
            return
        if isinstance(self.result, AnalysisResult):
            if not self.operation_options[-1].can_output:
                return
            self.operation_options[-1].analysis_output_func(self.result, self.operation_options[-1].out_path)
            return
        if not isinstance(self.result, DataSource):
            raise NotImplementedError(f'have not implemented pipeline output for type {type(self.result)}')
        if self.result.location is None:
            if not self.operation_options[-1].can_output:
                return
            self.result.location = self.operation_options[-1].out_path

        # By default, save calculated variables, unless user explicitly passes to not save them
        # Essentially setting the opposite default versus working directly with the DataSource since
        # usually DataSource calculations are done on loading and it is assumed if the pipeline result
        # is being saved at all then it is likely an expensive calculation which the user doesn't
        # want to repeat on every load
        if 'save_calculated' not in self.result.data_outputter_kwargs:
            extra_kwargs = dict(save_calculated=True)
        else:
            extra_kwargs = {}

        self.result.output(**extra_kwargs)

    def summary(self, *summary_args, summary_method: str=None, summary_function: Callable=None,
                             summary_attr: str=None, **summary_method_kwargs):
        for op in self.operations:
            op.summary(
                *summary_args,
                summary_method=summary_method,
                summary_function=summary_function,
                summary_attr=summary_attr,
                **summary_method_kwargs
            )

    def describe(self):
        for op in self.operations:
            op.describe()

    @property
    def last_modified(self) -> Optional[datetime.datetime]:
        if any([obj.last_modified is None for obj in self.operations]):
            return None

        return max([source.last_modified for source in self._objs_with_last_modified])

    @property
    def obj_last_modified(self) -> Optional[ObjWithLastModified]:
        if not self._objs_with_last_modified:
            return None
        most_recent_time = datetime.datetime(1900, 1, 1)
        most_recent_index = None
        for i, obj in enumerate(self._objs_with_last_modified):
            if obj.last_modified > most_recent_time:
                most_recent_time = obj.last_modified
                most_recent_index = i

        if most_recent_index is not None:
            return self._objs_with_last_modified[most_recent_index]

    @property
    def _objs_with_last_modified(self) -> List[ObjWithLastModified]:
        objs_with_last_modified: List[ObjWithLastModified]
        objs_with_last_modified = self.operations
        objs_with_last_modified = [obj for obj in objs_with_last_modified if obj.last_modified is not None]
        return objs_with_last_modified

    @property
    def allow_modifying_result(self) -> bool:
        return self.operation_options[-1].allow_modifying_result

    def copy(self):
        return deepcopy(self)

    @property
    def _graph_contents(self) -> Sequence[GraphObject]:
        elems = [self.primary_node]
        for source in self.data_sources:
            elems.extend(source._graph_contents)
            edge = Edge(source.primary_node, self.primary_node)
            elems.append(edge)
        return elems

    @property
    def graph(self) -> Digraph:
        elems = self._graph_contents
        graph = Digraph(self.name)
        for elem in elems:
            elem.add_to_graph(graph)
        return graph


class LastOperationFinishedException(Exception):
    pass


def _get_operations_for_single(data_source: DataSourceOrPipeline, options: OperationOptions) -> List[DataOperation]:
    """
     Creates a list of DataOperation/subclass objects from a single DataSource or DataPipeline object
    :param data_source:
    :param options: Options for the main operation
    :return:
    """
    operations: List[DataOperation] = []
    final_operation_sources: List[DataSource] = []
    # Add any pipeline operations first, as the results from the pipeline must be ready before we can use the results
    # for other data sources or pipeline results operations
    if _is_data_pipeline(data_source):
        operations += data_source.operations  # type: ignore
        pipeline_result = data_source.operations[-1].result  # type: ignore
        # result of first pipeline will be first source in final operation
        final_operation_sources.append(pipeline_result)
    else:
        final_operation_sources.append(data_source)  # type: ignore

    # Add last (or only) operation
    operations.append(options.get_operation(final_operation_sources, options))

    return operations


def _get_operations_for_pair(data_source_1: DataSourceOrPipeline, data_source_2: DataSourceOrPipeline,
                             options: OperationOptions) -> List[DataOperation]:
    """
    Creates a list of DataOperation/subclass objects from a paring of two DataSource objects, a DataSource and a
    DataPipeline, or two DataPipeline objects.
    :param data_source_1: DataSource or DataMergePipeline
    :param data_source_2: DataSource or DataMergePipeline
    :param options: Options for the main operation
    :return: list of DataOperation/subclass objects
    """
    operations: List[DataOperation] = []
    final_operation_sources: List[DataSource] = []
    # Add any pipeline operations first, as the results from the pipeline must be ready before we can use the results
    # for other data sources or pipeline results operations
    if _is_data_pipeline(data_source_1):
        operations += data_source_1.operations  # type: ignore
        pipeline_1_result = data_source_1.operations[-1].result  # type: ignore
        # result of first pipeline will be first source in final operation
        final_operation_sources.append(pipeline_1_result)
    else:
        final_operation_sources.append(data_source_1)  # type: ignore

    if _is_data_pipeline(data_source_2):
        operations += data_source_2.operations  # type: ignore
        # result of second pipeline will be second source in final operation
        pipeline_2_result = data_source_2.operations[-1].result # type: ignore
        final_operation_sources.append(pipeline_2_result)
    else:
        final_operation_sources.append(data_source_2)

    # Add last (or only) operation
    operations.append(options.get_operation(final_operation_sources, options))

    return operations


def _is_data_pipeline(obj) -> bool:
    return hasattr(obj, 'data_sources') and hasattr(obj, 'operation_options')


import datetime
from copy import deepcopy
from typing import Sequence, Callable, Optional, Type, TYPE_CHECKING, Dict, Any

from mixins.repr import ReprMixin

if TYPE_CHECKING:
    from datacode.models.analysis import AnalysisResult

from datacode.models.source import DataSource
import datacode.hooks as hooks


class DataOperation(ReprMixin):
    """
    Base class for a singlar data process that takes one or more DataSources as inputs and has one DataSource
    as the output.
    """
    repr_cols = ['options', 'output_name', 'data_sources']
    num_required_sources: int = 1

    def __init__(self, data_sources: Sequence[DataSource], options: 'OperationOptions',
                 output_name: Optional[str] = None, **result_kwargs):
        if output_name is None:
            names = [ds.name if ds.name is not None else 'unnamed' for ds in data_sources]
            output_name = ' & '.join(names) + ' Post-Operation'
        self.options = options
        self.data_sources = data_sources
        self.output_name = output_name
        self.result = None
        self.result_kwargs = result_kwargs
        self._set_result(**result_kwargs)

    def execute(self):
        hooks.on_begin_execute_operation(self)
        self._execute()
        hooks.on_end_execute_operation(self)

    def _execute(self):
        raise NotImplementedError('must implement _execute in subclass of DataOperation')

    def summary(self, *summary_args, summary_method: str=None, summary_function: Callable=None,
                             summary_attr: str=None, **summary_method_kwargs):
        raise NotImplementedError('must implement summary in subclass of DataOperation')

    def describe(self):
        raise NotImplementedError('must implement describe in subclass of DataOperation')

    @property
    def last_modified(self) -> Optional[datetime.datetime]:
        if self.options.last_modified is not None or not self.data_sources:
            return self.options.last_modified

        if any([ds.last_modified is None for ds in self.data_sources]):
            return None

        last_modified = max([ds.last_modified for ds in self.data_sources])
        return last_modified

    def _set_result(self, **kwargs):
        self.result = self.options.result_class(name=self.output_name, location=self.options.out_path, **kwargs)

    def __repr__(self):
        return f'<DataOperation(data_sources={self.data_sources}, result={self.result})>'


class OperationOptions(ReprMixin):
    """
    Base class for options passed to DataOperations
    """
    op_class: Type[DataOperation] = DataOperation
    result_class: Type = DataSource
    result_kwargs: Optional[Dict[str, Any]] = None
    out_path: Optional[str] = None
    last_modified: Optional[datetime.datetime] = None
    allow_modifying_result: bool = True
    analysis_output_func: Optional[Callable[['AnalysisResult', str], None]] = None
    repr_cols = [
        'out_path',
        'last_modified',
        'allow_modifying_result',
        'analysis_output_func'
    ]

    def copy(self):
        return deepcopy(self)

    @property
    def can_output(self) -> bool:
        return self.out_path is not None

    def get_operation(self, *args) -> DataOperation:
        kwargs = {}
        if self.result_kwargs is not None:
            kwargs = self.result_kwargs
        return self.op_class(*args, **kwargs)



from typing import Sequence, Optional

import pandas as pd
from pyexlatex import Document

from datacode.typing import (
    StrOrNone,
    DocumentOrLatexObjs,
    IntOrNone,
    FloatOrNone,
    IntSequenceOrNone,
    FloatSequenceOrNone
)
from datacode.summarize.subset.missing.summary.graphs.pctbyid import missing_pct_by_id_figure
from datacode.summarize.subset.missing.detail.main import obs_and_id_count_and_missing_pct_table

def missing_data_single_column_analysis(df: pd.DataFrame, col_with_missings: str, id_col: str,
                                        row_byvar: str, col_byvar: str, datevar: str,
                                        missing_tolerance: IntOrNone = 0, missing_quantile: FloatOrNone = None,
                                        summary_missing_tolerances: IntSequenceOrNone = (0, 5, 10),
                                        summary_missing_quantiles: FloatSequenceOrNone = None,
                                        sort_cols_as_numeric: bool = True, sort_rows_as_numeric: bool = True,
                                        sort_cols_as_portvar: bool=False, sort_rows_as_portvar: bool=False,
                                        count_format_str: str = '.0f', pct_format_str: str = '.1f',
                                        missing_display_str: str = 'Missing', period_display_name: str='Period',
                                        extra_caption: str='', table_extra_below_text: str='',
                                        table_align: str=None,
                                        outfolder: StrOrNone=None, as_document: bool=True,
                                        document_title: StrOrNone=None, authors: Optional[Sequence[str]] = None,
                                        ) -> DocumentOrLatexObjs:

    missing_pct_fig = missing_pct_by_id_figure(
        df,
        id_col,
        col_with_missings,
        outfolder=outfolder,
        outname=f'Percentage {missing_display_str} Obs by Firm for {col_with_missings}'
    )

    missing_pct_table = obs_and_id_count_and_missing_pct_table(
        df,
        col_with_missings,
        id_col,
        row_byvar,
        col_byvar,
        datevar,
        missing_tolerance=missing_tolerance,
        missing_quantile=missing_quantile,
        summary_missing_tolerances=summary_missing_tolerances,
        summary_missing_quantiles=summary_missing_quantiles,
        sort_cols_as_numeric=sort_cols_as_numeric,
        sort_rows_as_numeric=sort_rows_as_numeric,
        sort_cols_as_portvar=sort_cols_as_portvar,
        sort_rows_as_portvar=sort_rows_as_portvar,
        count_format_str=count_format_str,
        pct_format_str=pct_format_str,
        missing_display_str=missing_display_str,
        period_display_name=period_display_name,
        extra_caption=extra_caption,
        extra_below_text=table_extra_below_text,
        table_align=table_align,
        outfolder=outfolder
    )

    if not as_document:
        return [missing_pct_fig, missing_pct_table]

    if document_title is None:
        document_title = f'{missing_display_str} {col_with_missings} Analysis'

    if extra_caption:
        document_title = f'{document_title} - {extra_caption}'

    doc = Document(
        [missing_pct_fig, missing_pct_table],
        title=document_title,
        authors=authors
    )

    if outfolder:
        doc.to_pdf(
            outname=document_title,
            outfolder=outfolder
        )

    return doc